import os

import pycardano
import sys

import subprocess

import tempfile

import unittest

import frozendict
import frozenlist2
import hypothesis
from hypothesis import given
from hypothesis import strategies as st
from opshin import IndefiniteList
from parameterized import parameterized

from uplc import ast as uplc, eval as uplc_eval

from . import PLUTUS_VM_PROFILE
from .. import prelude, builder, Purpose, PlutusContract
from .utils import eval_uplc_value, Unit, eval_uplc
from ..bridge import wraps_builtin
from ..compiler_config import OPT_O2_CONFIG, DEFAULT_CONFIG

hypothesis.settings.load_profile(PLUTUS_VM_PROFILE)

# these imports are required to eval the result of script context dumps
from ..ledger.api_v2 import *
from pycardano import RawPlutusData, RawCBOR
from cbor2 import CBORTag

DEFAULT_CONFIG_FORCE_THREE_PARAMS = DEFAULT_CONFIG.update(force_three_params=True)
DEFAULT_CONFIG_CONSTANT_FOLDING = DEFAULT_CONFIG.update(constant_folding=True)

ALL_EXAMPLES = [
    os.path.join(root, f)
    for root, dirs, files in os.walk("examples")
    for f in files
    if f.endswith(".py") and not f.startswith("broken") and not f.startswith("extract")
]


def fib(n):
    a, b = 0, 1
    for _ in range(n):
        a, b = b, a + b
    return a


@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int


@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: int
    bar: int


a_or_b = st.sampled_from([A(0), B(1, 2)])
some_output = st.sampled_from([SomeOutputDatum(b"0"), SomeOutputDatumHash(b"1")])


class MiscTest(unittest.TestCase):
    def test_assert_sum_contract_succeed(self):
        input_file = "examples/smart_contracts/assert_sum.py"
        with open(input_file) as fp:
            source_code = fp.read()
        ret = eval_uplc(
            source_code,
            20,
            22,
            uplc.data_from_cbor(
                bytes.fromhex(
                    "d8799fd8799f9fd8799fd8799fd8799f582055d353acacaab6460b37ed0f0e3a1a0aabf056df4a7fa1e265d21149ccacc527ff01ffd8799fd8799fd87a9f581cdbe769758f26efb21f008dc097bb194cffc622acc37fcefc5372eee3ffd87a80ffa140a1401a00989680d87a9f5820dfab81872ce2bbe6ee5af9bbfee4047f91c1f57db5e30da727d5fef1e7f02f4dffd87a80ffffff809fd8799fd8799fd8799f581cdc315c289fee4484eda07038393f21dc4e572aff292d7926018725c2ffd87a80ffa140a14000d87980d87a80ffffa140a14000a140a1400080a0d8799fd8799fd87980d87a80ffd8799fd87b80d87a80ffff80a1d87a9fd8799fd8799f582055d353acacaab6460b37ed0f0e3a1a0aabf056df4a7fa1e265d21149ccacc527ff01ffffd87980a15820dfab81872ce2bbe6ee5af9bbfee4047f91c1f57db5e30da727d5fef1e7f02f4dd8799f581cdc315c289fee4484eda07038393f21dc4e572aff292d7926018725c2ffd8799f5820746957f0eb57f2b11119684e611a98f373afea93473fefbb7632d579af2f6259ffffd87a9fd8799fd8799f582055d353acacaab6460b37ed0f0e3a1a0aabf056df4a7fa1e265d21149ccacc527ff01ffffff"
                )
            ),
            config=DEFAULT_CONFIG.update(OPT_O2_CONFIG),
        )
        self.assertEqual(ret, uplc.PlutusConstr(0, []))

    @unittest.expectedFailure
    def test_assert_sum_contract_fail(self):
        input_file = "examples/smart_contracts/assert_sum.py"
        with open(input_file) as fp:
            source_code = fp.read()
        ret = eval_uplc(source_code, 0, 22, Unit())

    @given(
        a=st.integers(min_value=-10, max_value=10),
        b=st.integers(min_value=0, max_value=10),
    )
    def test_mult_for(self, a: int, b: int):
        input_file = "examples/mult_for.py"
        with open(input_file) as fp:
            source_code = fp.read()
        ret = eval_uplc_value(source_code, a, b)
        self.assertEqual(ret, a * b)

    @given(
        a=st.integers(min_value=-10, max_value=10),
        b=st.integers(min_value=0, max_value=10),
    )
    def test_mult_for(self, a: int, b: int):
        input_file = "examples/mult_for.py"
        with open(input_file) as fp:
            source_code = fp.read()
        ret = eval_uplc_value(source_code, a, b)
        self.assertEqual(ret, a * b)

    @given(
        a=st.integers(min_value=-10, max_value=10),
        b=st.integers(min_value=0, max_value=10),
    )
    def test_mult_for_return(self, a: int, b: int):
        source_code = """
def validator(a: int, b: int) -> int:
    c = 0
    i = 0
    for i in range(b):
        c += a
        if i == 1:
            return c
    return c
"""
        ret = eval_uplc_value(source_code, a, b)
        self.assertEqual(ret, a * min(b, 2))

    @given(
        a=st.integers(min_value=-10, max_value=10),
        b=st.integers(min_value=0, max_value=10),
    )
    def test_mult_while_return(self, a: int, b: int):
        source_code = """
def validator(a: int, b: int) -> int:
    c = 0
    i = 0
    while i < b:
        c += a
        i += 1
        if i == 2:
            return c
    return c
"""
        ret = eval_uplc_value(source_code, a, b)
        self.assertEqual(ret, a * min(2, b))

    @given(
        a=st.integers(),
        b=st.integers(),
    )
    def test_sum(self, a: int, b: int):
        input_file = "examples/sum.py"
        with open(input_file) as fp:
            source_code = fp.read()
        ret = eval_uplc_value(source_code, a, b)
        self.assertEqual(ret, a + b)

    def test_complex_datum_correct_vals(self):
        input_file = "examples/complex_datum.py"
        with open(input_file) as fp:
            source_code = fp.read()
        ret = eval_uplc_value(
            source_code,
            uplc.data_from_cbor(
                bytes.fromhex(
                    "d8799fd8799fd8799f581c81aab0790f33d26bad68a6a13ae98562aa1366da48cdce20dec21acfffd8799fd8799fd8799f581c145db8343296bd214dde862a64d700c29ed8a71d58bcf865659f5463ffffffffd8799fd8799f581c81aab0790f33d26bad68a6a13ae98562aa1366da48cdce20dec21acfffd8799fd8799fd8799f581c145db8343296bd214dde862a64d700c29ed8a71d58bcf865659f5463ffffffffd87a80d8799f1a38220b0bff1a001e84801a001e8480582051176daeee7f2ce62963c50a16f641951e21b8522da262980d4dd361a9bf331b4e4d7565736c69537761705f414d4dff"
                )
            ),
        )
        self.assertEqual(
            bytes.fromhex("81aab0790f33d26bad68a6a13ae98562aa1366da48cdce20dec21acf"),
            ret,
        )

    def test_hello_world(self):
        input_file = "examples/hello_world.py"
        with open(input_file) as fp:
            source_code = fp.read()
        ret = eval_uplc(source_code, Unit())

    def test_list_datum_correct_vals(self):
        input_file = "examples/list_datum.py"
        with open(input_file) as fp:
            source_code = fp.read()
        ret = eval_uplc_value(
            source_code, uplc.data_from_cbor(bytes.fromhex("d8799f9f41014102ffff"))
        )
        self.assertEqual(
            1,
            ret,
        )

    def test_showcase(self):
        input_file = "examples/showcase.py"
        with open(input_file) as fp:
            source_code = fp.read()
        ret = eval_uplc_value(source_code, 1)
        self.assertEqual(
            42,
            ret,
        )

    @given(n=st.integers(min_value=0, max_value=5))
    def test_fib_iter(self, n):
        input_file = "examples/fib_iter.py"
        with open(input_file) as fp:
            source_code = fp.read()
        ret = eval_uplc_value(source_code, n)
        self.assertEqual(
            fib(n),
            ret,
        )

    @given(n=st.integers(min_value=0, max_value=5))
    def test_fib_rec(self, n):
        input_file = "examples/fib_rec.py"
        with open(input_file) as fp:
            source_code = fp.read()
        ret = eval_uplc_value(source_code, n)
        self.assertEqual(
            fib(n),
            ret,
        )

    def test_gift_contract_succeed(self):
        input_file = "examples/smart_contracts/gift.py"
        with open(input_file) as fp:
            source_code = fp.read()
        ret = eval_uplc(
            source_code,
            uplc.PlutusConstr(
                0,
                [
                    uplc.PlutusByteString(
                        bytes.fromhex(
                            "dc315c289fee4484eda07038393f21dc4e572aff292d7926018725c2"
                        )
                    )
                ],
            ),
            uplc.PlutusConstr(0, []),
            uplc.data_from_cbor(
                bytes.fromhex(
                    (
                        "d8799fd8799f9fd8799fd8799fd8799f582055d353acacaab6460b37ed0f0e3a1a0aabf056df4a7fa1e265d21149ccacc527ff01ffd8799fd8799fd87a9f581cdbe769758f26efb21f008dc097bb194cffc622acc37fcefc5372eee3ffd87a80ffa140a1401a00989680d87a9f5820dfab81872ce2bbe6ee5af9bbfee4047f91c1f57db5e30da727d5fef1e7f02f4dffd87a80ffffff809fd8799fd8799fd8799f581cdc315c289fee4484eda07038393f21dc4e572aff292d7926018725c2ffd87a80ffa140a14000d87980d87a80ffd8799fd8799fd8799f581cdc315c289fee4484eda07038393f21dc4e572aff292d7926018725c2ffd87a80ffa140a1401a000f4240d87980d87a80ffffa140a14000a140a1400080a0d8799fd8799fd87a9f1b000001836ac117d8ffd87a80ffd8799fd87b80d87a80ffff9f581cdc315c289fee4484eda07038393f21dc4e572aff292d7926018725c2ffa1d87a9fd8799fd8799f582055d353acacaab6460b37ed0f0e3a1a0aabf056df4a7fa1e265d21149ccacc527ff01ffffd87980a15820dfab81872ce2bbe6ee5af9bbfee4047f91c1f57db5e30da727d5fef1e7f02f4dd8799f581cdc315c289fee4484eda07038393f21dc4e572aff292d7926018725c2ffd8799f5820c17c32f6433ae22c2acaebfb796bbfaee3993ff7ebb58a2bac6b4a3bdd2f6d28ffffd87a9fd8799fd8799f582055d353acacaab6460b37ed0f0e3a1a0aabf056df4a7fa1e265d21149ccacc527ff01ffffff"
                    )
                )
            ),
        )
        self.assertEqual(ret, uplc.PlutusConstr(0, []))

    @unittest.expectedFailure
    def test_gift_contract_fail(self):
        input_file = "examples/smart_contracts/gift.py"
        with open(input_file) as fp:
            source_code = fp.read()
        # required sig missing int this script context
        ret = eval_uplc(
            source_code,
            uplc.PlutusConstr(
                0,
                [
                    uplc.PlutusByteString(
                        bytes.fromhex(
                            "dc315c289fee4484eda07038393f21dc4e572aff292d7926018725c2"
                        )
                    )
                ],
            ),
            uplc.PlutusConstr(0, []),
            uplc.data_from_cbor(
                bytes.fromhex(
                    (
                        "d8799fd8799f9fd8799fd8799fd8799f582055d353acacaab6460b37ed0f0e3a1a0aabf056df4a7fa1e265d21149ccacc527ff01ffd8799fd8799fd87a9f581cdbe769758f26efb21f008dc097bb194cffc622acc37fcefc5372eee3ffd87a80ffa140a1401a00989680d87a9f5820dfab81872ce2bbe6ee5af9bbfee4047f91c1f57db5e30da727d5fef1e7f02f4dffd87a80ffffff809fd8799fd8799fd8799f581cdc315c289fee4484eda07038393f21dc4e572aff292d7926018725c2ffd87a80ffa140a14000d87980d87a80ffffa140a14000a140a1400080a0d8799fd8799fd87a9f1b000001836ac117d8ffd87a80ffd8799fd87b80d87a80ffff80a1d87a9fd8799fd8799f582055d353acacaab6460b37ed0f0e3a1a0aabf056df4a7fa1e265d21149ccacc527ff01ffffd87980a15820dfab81872ce2bbe6ee5af9bbfee4047f91c1f57db5e30da727d5fef1e7f02f4dd8799f581cdc315c289fee4484eda07038393f21dc4e572aff292d7926018725c2ffd8799f5820797a1e1720b63621c6b185088184cb8e23af6e46b55bd83e7a91024c823a6c2affffd87a9fd8799fd8799f582055d353acacaab6460b37ed0f0e3a1a0aabf056df4a7fa1e265d21149ccacc527ff01ffffff"
                    )
                )
            ),
        )

    def test_recursion_simple(self):
        source_code = """
def validator(_: None) -> int:
    def a(n: int) -> int:
      if n == 0:
        res = 0
      else:
        res = a(n-1)
      return res
    return a(1)
        """
        ret = eval_uplc_value(source_code, Unit())
        self.assertEqual(0, ret)

    @unittest.expectedFailure
    def test_recursion_illegal(self):
        # this is now an illegal retyping because read variables dont match
        source_code = """
def validator(_: None) -> int:
    def a(n: int) -> int:
      if n == 0:
        res = 0
      else:
        res = a(n-1)
      return res
    b = a
    def a(x: int) -> int:
      return 100
    return b(1)
        """
        ret = eval_uplc_value(source_code, Unit())
        self.assertEqual(100, ret)

    def test_recursion_legal(self):
        source_code = """
def validator(_: None) -> int:
    def a(n: int) -> int:
      if n == 0:
        res = 0
      else:
        res = a(n-1)
      return res
    b = a
    def a(n: int) -> int:
      a
      if 1 == n:
        pass
      return 100
    return b(1)
        """
        ret = eval_uplc_value(source_code, Unit())
        self.assertEqual(100, ret)

    @unittest.expectedFailure
    def test_uninitialized_access(self):
        source_code = """
def validator(_: None) -> int:
    b = 1
    def a(n: int) -> int:
      b += 1
      if b == 2:
        return 0
      else:
        return a(n-1)
    return a(1)
        """
        builder._compile(source_code)

    @unittest.expectedFailure
    def test_illegal_bind(self):
        source_code = """
def validator(_: None) -> int:
    b = 1
    def a(n: int) -> int:
      if n == 0:
        return 100
      if b == 2:
        return 0
      b = 2
      return a(n-1)
    return a(2)
        """
        builder._compile(source_code)

    @unittest.expectedFailure
    def test_type_reassignment_function_bound(self):
        # changing the type of a variable should be disallowed if the variable is bound by a function
        # it can be ok if the types can be merged (resulting in union type inside the function) but
        # generally should be disallowed
        source_code = """
def validator(_: None) -> int:
    b = 1
    def a(n: int) -> int:
      return b
    b = b''
    return a(1)
        """
        builder._compile(source_code)

    @unittest.expectedFailure
    def test_illegal_function_retype(self):
        source_code = """
def validator(_: None) -> int:
    def a(n: int) -> int:
      if n == 0:
        res = 0
      else:
        res = a(n-1)
      return res
    b = a
    def a() -> int:
      return 100
    return b(1)
        """
        builder._compile(source_code)

    def test_datum_cast(self):
        input_file = "examples/datum_cast.py"
        with open(input_file) as fp:
            source_code = fp.read()
        ret = eval_uplc_value(
            source_code,
            uplc.data_from_cbor(
                bytes.fromhex(
                    "d8799fd8799fd8799f581c81aab0790f33d26bad68a6a13ae98562aa1366da48cdce20dec21acfffd8799fd8799fd8799f581c145db8343296bd214dde862a64d700c29ed8a71d58bcf865659f5463ffffffffd8799fd8799f581c81aab0790f33d26bad68a6a13ae98562aa1366da48cdce20dec21acfffd8799fd8799fd8799f581c145db8343296bd214dde862a64d700c29ed8a71d58bcf865659f5463ffffffffd87a80d8799f1a38220b0bff1a001e84801a001e8480582051176daeee7f2ce62963c50a16f641951e21b8522da262980d4dd361a9bf331b4e4d7565736c69537761705f414d4dff"
                )
            ),
            uplc.PlutusByteString(b"test"),
        )
        self.assertEqual(
            bytes.fromhex("81aab0790f33d26bad68a6a13ae98562aa1366da48cdce20dec21acf")
            + b"test",
            ret,
        )

    def test_wrapping_contract_compile(self):
        # TODO devise tests for this
        input_file = "examples/smart_contracts/wrapped_token.py"
        with open(input_file) as fp:
            source_code = fp.read()
        builder._compile(source_code, config=DEFAULT_CONFIG_FORCE_THREE_PARAMS)

    def test_dual_use_compile(self):
        # TODO devise tests for this
        input_file = "examples/smart_contracts/dual_use.py"
        with open(input_file) as fp:
            source_code = fp.read()
        builder._compile(source_code, config=DEFAULT_CONFIG_FORCE_THREE_PARAMS)

    def test_marketplace_compile(self):
        # TODO devise tests for this
        input_file = "examples/smart_contracts/marketplace.py"
        with open(input_file) as fp:
            source_code = fp.read()
        builder._compile(source_code)

    @unittest.expectedFailure
    def test_marketplace_compile_fail(self):
        input_file = "examples/smart_contracts/marketplace.py"
        with open(input_file) as fp:
            source_code = fp.read()
        builder._compile(source_code, config=DEFAULT_CONFIG_FORCE_THREE_PARAMS)

    def test_parameterized_compile(self):
        # TODO devise tests for this
        input_file = "examples/smart_contracts/parameterized.py"
        with open(input_file) as fp:
            source_code = fp.read()
        builder._compile(source_code, config=DEFAULT_CONFIG_FORCE_THREE_PARAMS)

    def test_dict_datum(self):
        input_file = "examples/dict_datum.py"
        with open(input_file) as fp:
            source_code = fp.read()
        d = uplc.PlutusConstr(
            0,
            [
                uplc.PlutusMap(
                    frozendict.frozendict(
                        {
                            uplc.PlutusConstr(
                                0,
                                frozenlist2.frozenlist(
                                    [uplc.PlutusByteString(b"\x01")]
                                ),
                            ): uplc.PlutusInteger(2)
                        }
                    )
                )
            ],
        )
        ret = eval_uplc(source_code, d)
        self.assertTrue(bool(ret))

    def test_dict_datum_wrong(self):
        input_file = "examples/dict_datum.py"
        with open(input_file) as fp:
            source_code = fp.read()
        d = uplc.PlutusConstr(
            0,
            [
                uplc.PlutusMap(
                    frozendict.frozendict(
                        {
                            uplc.PlutusConstr(
                                0,
                                frozenlist2.frozenlist(
                                    [uplc.PlutusByteString(b"\x02")]
                                ),
                            ): uplc.PlutusInteger(2)
                        }
                    )
                )
            ],
        )
        ret = eval_uplc_value(source_code, d)
        self.assertFalse(bool(ret))

    def test_removedeadvar_noissue(self):
        source_code = """
from opshin.prelude import *
def validator(x: Token) -> bool:
    b = 4
    a = b
    return True
        """
        ret = eval_uplc_value(source_code, Unit())
        self.assertEqual(ret, True)

    def test_removedeadvar_noissue2(self):
        source_code = """
from opshin.prelude import *
def validator(x: Token) -> bool:
    def foo(x: Token) -> bool:
        b = 4
        a = b
        return True
    return foo(x)
        """
        ret = eval_uplc_value(source_code, Unit())
        self.assertEqual(ret, True)

    def test_removedeadvar_noissue3(self):
        source_code = """
from opshin.prelude import *

def foo(x: Token) -> bool:
    b = 4
    a = b
    return True

def validator(x: Token) -> bool:
    return foo(x)
        """
        ret = eval_uplc_value(source_code, Unit())
        self.assertEqual(ret, True)

    @unittest.expectedFailure
    def test_overopt_removedeadvar(self):
        source_code = """
from opshin.prelude import *
def validator(x: Token) -> bool:
    a = x.policy_id
    return True
        """
        ret = eval_uplc(source_code, Unit())

    @unittest.expectedFailure
    def test_opt_shared_var(self):
        source_code = """
from opshin.prelude import *
def validator(x: Token) -> bool:
    if False:
        y = x
    else:
        a = y
    return True
        """
        ret = eval_uplc(source_code, Unit())

    def test_list_expr(self):
        # this tests that the list expression is evaluated correctly
        source_code = """
def validator(x: None) -> List[int]:
    return [1, 2, 3, 4, 5]
        """
        ret = eval_uplc_value(source_code, Unit())
        ret = [x.value for x in ret]
        self.assertEqual(ret, [1, 2, 3, 4, 5], "List expression incorrectly compiled")

    def test_list_expr_not_const(self):
        # this tests that the list expression is evaluated correctly (for non-constant expressions)
        source_code = """
def validator(x: int) -> List[int]:
    return [x, x+1, x+2, x+3, x+4]
        """
        ret = eval_uplc_value(source_code, 1)
        ret = [x.value for x in ret]
        self.assertEqual(ret, [1, 2, 3, 4, 5], "List expression incorrectly compiled")

    def test_dict_expr_not_const(self):
        # this tests that the list expression is evaluated correctly (for non-constant expressions)
        source_code = """
def validator(x: int) -> Dict[int, bytes]:
    return {x: b"a", x+1: b"b"}
        """
        ret = eval_uplc_value(source_code, 1)
        ret = {x.value: y.value for x, y in ret.items()}
        self.assertEqual(
            ret, {1: b"a", 2: b"b"}, "Dict expression incorrectly compiled"
        )

    def test_redefine_poly_constr(self):
        # this tests that classes defined by assignment inherit constructors
        source_code = """
def validator(x: None) -> bytes:
    a = bytes
    return a([2, 3])
        """
        ret = eval_uplc_value(source_code, Unit())
        self.assertEqual(ret, bytes([2, 3]), "Re-assignment of global variable failed")

    @given(st.booleans())
    def test_redefine_constr(self, x):
        # this tests that classes defined by assignment inherit constructors
        source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int
    bar: int
    
def validator(x: int) -> int:
    a = A
    return a(x, 1).foo
        """
        ret = eval_uplc_value(source_code, int(x))
        self.assertEqual(ret, int(x), "Re-assignment of class constr failed")

    def test_wrap_into_generic_data(self):
        # this tests data is wrapped into Anything if a function accepts Anything
        source_code = """
from opshin.prelude import *
def validator(_: None) -> SomeOutputDatum:
    return SomeOutputDatum(b"a")
        """
        ret = eval_uplc(source_code, Unit())
        self.assertEqual(
            ret,
            uplc.data_from_cbor(prelude.SomeOutputDatum(b"a").to_cbor()),
            "Wrapping to generic data failed",
        )

    def test_list_comprehension_even(self):
        input_file = "examples/list_comprehensions.py"
        with open(input_file) as fp:
            source_code = fp.read()
        ret = eval_uplc_value(source_code, 8, 1)
        ret = [x.value for x in ret]
        self.assertEqual(
            ret,
            [x * x for x in range(8) if x % 2 == 0],
            "List comprehension with filter incorrectly evaluated",
        )

    def test_list_comprehension_all(self):
        input_file = "examples/list_comprehensions.py"
        with open(input_file) as fp:
            source_code = fp.read()
        ret = eval_uplc_value(source_code, 8, 0)
        ret = [x.value for x in ret]
        self.assertEqual(
            ret,
            [x * x for x in range(8)],
            "List comprehension incorrectly evaluated",
        )

    def test_dict_comprehension_even(self):
        input_file = "examples/dict_comprehensions.py"
        with open(input_file) as fp:
            source_code = fp.read()
        ret = eval_uplc_value(source_code, 8, 1)
        ret = {x.value: y.value for x, y in ret.items()}
        self.assertEqual(
            ret,
            {x: x * x for x in range(8) if x % 2 == 0},
            "Dict comprehension incorrectly evaluated",
        )

    def test_dict_comprehension_all(self):
        input_file = "examples/dict_comprehensions.py"
        with open(input_file) as fp:
            source_code = fp.read()
        ret = eval_uplc_value(source_code, 8, 0)
        ret = {x.value: y.value for x, y in ret.items()}
        self.assertEqual(
            ret,
            {x: x * x for x in range(8)},
            "Dict comprehension incorrectly evaluated",
        )

    @hypothesis.given(some_output)
    def test_union_type_attr_access_all_records(self, x):
        source_code = """
from opshin.prelude import *

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: SomeOutputDatumHash
    
@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foo: SomeOutputDatum

def validator(x: Union[A, B]) -> Union[SomeOutputDatumHash, SomeOutputDatum]:
    return x.foo
"""

        @dataclass()
        class A(PlutusData):
            CONSTR_ID = 0
            foo: SomeOutputDatumHash

        @dataclass()
        class B(PlutusData):
            CONSTR_ID = 1
            foo: SomeOutputDatum

        x = A(x) if isinstance(x, SomeOutputDatumHash) else B(x)

        ret = eval_uplc(source_code, x)
        self.assertEqual(ret, uplc.data_from_cbor(x.foo.to_cbor()))

    @hypothesis.given(some_output, st.sampled_from([1, 2, 3]))
    def test_union_type_attr_access_all_records_diff_pos(self, x, y):
        source_code = """
from opshin.prelude import *

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: SomeOutputDatumHash

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foo: SomeOutputDatum
    
@dataclass()
class C(PlutusData):
    CONSTR_ID = 2
    bar: int
    foo: SomeOutputDatum
    
@dataclass()
class D(PlutusData):
    CONSTR_ID = 3
    foobar: int
    bar: int
    foo: SomeOutputDatum

def validator(x: Union[A, B, C, D]) -> Union[SomeOutputDatumHash, SomeOutputDatum]:
    return x.foo
"""

        @dataclass()
        class A(PlutusData):
            CONSTR_ID = 0
            foo: SomeOutputDatumHash

        @dataclass()
        class B(PlutusData):
            CONSTR_ID = 1
            foo: SomeOutputDatum

        @dataclass()
        class C(PlutusData):
            CONSTR_ID = 2
            bar: int
            foo: SomeOutputDatum

        @dataclass()
        class D(PlutusData):
            CONSTR_ID = 3
            foobar: int
            bar: int
            foo: SomeOutputDatum

        x = (
            A(x)
            if isinstance(x, SomeOutputDatumHash)
            else B(x) if y == 1 else C(0, x) if y == 2 else D(0, 0, x)
        )

        ret = eval_uplc(source_code, x)
        self.assertEqual(ret, uplc.data_from_cbor(x.foo.to_cbor()))

    @unittest.expectedFailure
    def test_union_type_all_records_same_constr(self):
        source_code = """
from opshin.prelude import *

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: SomeOutputDatumHash

@dataclass()
class B(PlutusData):
    CONSTR_ID = 0
    foo: SomeOutputDatum

def validator(x: Union[A, B]) -> Union[SomeOutputDatumHash, SomeOutputDatum]:
    return x.foo
"""
        eval_uplc(source_code, Unit())

    @unittest.expectedFailure
    def test_union_type_attr_access_all_records_same_constr(self):
        source_code = """
from opshin.prelude import *

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: Token

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foo: Address

def validator(x: Union[A, B]) -> int:
    m = x.foo
    if isinstance(m, Address):
        k = 0
    else:
        k = 1
    return k
"""
        eval_uplc(source_code, Unit())

    def test_union_type_attr_access_maximum_type(self):
        source_code = """
from opshin.prelude import *

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foo: int

def validator(x: Union[A, B]) -> int:
    return x.foo
"""
        ret = eval_uplc_value(
            source_code, uplc.PlutusConstr(0, [uplc.PlutusInteger(1)])
        )
        self.assertEqual(ret, 1)

    def test_union_type_attr_anytype(self):
        source_code = """
from opshin.prelude import *

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: bytes

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foo: int

def validator(x: Union[A, B]) -> Anything:
    return x.foo
"""
        ret = eval_uplc_value(
            source_code, uplc.PlutusConstr(0, [uplc.PlutusByteString(b"")])
        )
        self.assertEqual(ret, b"")

    def test_typecast_anything_int(self):
        source_code = """
def validator(x: Anything) -> int:
    b: int = x
    return b
"""
        ret = eval_uplc_value(source_code, 0)
        self.assertEqual(ret, 0)

    def test_typecast_int_anything(self):
        # this should compile, it happens implicitly anyways when calling a function with Any parameters
        source_code = """
def validator(x: int) -> Anything:
    b: Anything = x
    return b
"""
        ret = eval_uplc_value(source_code, 0)
        self.assertEqual(ret, 0)

    def test_typecast_int_anything_int(self):
        source_code = """
def validator(x: int) -> Anything:
    b: Anything = x
    c: int = b
    return c + 1
"""
        ret = eval_uplc_value(source_code, 0)
        self.assertEqual(ret, 1)

    def test_typecast_anything_int_anything(self):
        source_code = """
def validator(x: Anything) -> Anything:
    b: int = x
    c: Anything = b + 1
    return c
"""
        ret = eval_uplc_value(source_code, 0)
        self.assertEqual(ret, 1)

    @unittest.expectedFailure
    def test_typecast_int_str(self):
        # this should not compile, the two types are unrelated and there is no meaningful way to cast them either direction
        source_code = """
def validator(x: int) -> str:
    b: str = x
    return b
"""
        builder._compile(source_code)

    def test_typecast_int_int(self):
        source_code = """
def validator(x: int) -> int:
    b: int = x
    return b
"""
        ret = eval_uplc_value(source_code, 0)
        self.assertEqual(ret, 0)

    def test_zero_ary(self):
        source_code = """
def a() -> None:
    assert False, "Executed a"

def validator(x: None) -> int:
    b = a
    if False:
        b()
    return 2
"""
        ret = eval_uplc_value(source_code, 0)
        self.assertEqual(ret, 2, "Invalid return value")

    @unittest.expectedFailure
    def test_zero_ary_exec(self):
        source_code = """
def a() -> None:
    assert False, "Executed a"

def validator(x: None) -> None:
    b = a
    if True:
        b()
"""
        ret = eval_uplc_value(source_code, 0)

    def test_zero_ary_method(self):
        source_code = """
def validator(x: None) -> None:
    b = b"\\xFF".decode
    if False:
        b()
"""
        eval_uplc(source_code, 0)

    @unittest.expectedFailure
    def test_zero_ary_method_exec(self):
        source_code = """
def validator(x: None) -> None:
    b = b"\\xFF".decode
    if True:
        b()
"""
        eval_uplc(source_code, 0)

    def test_zero_ary_method_exec_suc(self):
        source_code = """
def validator(x: None) -> str:
    b = b"\\x32".decode
    return b()
"""
        res = eval_uplc_value(source_code, 0)
        self.assertEqual(res, b"\x32")

    def test_return_anything(self):
        source_code = """
from opshin.prelude import *

def validator() -> Anything:
    return b""
"""
        res = eval_uplc(source_code, 0)
        self.assertEqual(res, uplc.PlutusByteString(b""))

    def test_no_return_annotation(self):
        source_code = """
from opshin.prelude import *

def validator():
    return b""
"""
        res = eval_uplc(source_code, 0)
        self.assertEqual(res, uplc.PlutusByteString(b""))

    def test_no_parameter_annotation(self):
        source_code = """
from opshin.prelude import *

def validator(a) -> bytes:
    b: bytes = a
    return b
"""
        res = eval_uplc(source_code, b"")
        self.assertEqual(res, uplc.PlutusByteString(b""))

    @given(xs=st.dictionaries(st.integers(), st.binary()))
    def test_dict_items_values_deconstr(self, xs):
        # asserts that deconstruction of parameters works for for loops too
        source_code = """
def validator(xs: Dict[int, bytes]) -> bytes:
    sum_values = b""
    for _, x in xs.items():
        sum_values += x
    return sum_values
"""
        ret = eval_uplc_value(source_code, xs)
        self.assertEqual(
            ret,
            b"".join(xs.values()),
            "for loop deconstruction did not behave as expected",
        )

    def test_nested_deconstruction(self):
        source_code = """
def validator(xs) -> int:
    a, ((b, c), d) = (1, ((2, 3), 4))
    return a + b + c + d
"""
        ret = eval_uplc_value(source_code, Unit())
        self.assertEqual(
            ret,
            1 + 2 + 3 + 4,
            "for loop deconstruction did not behave as expected",
        )

    @given(
        xs=st.dictionaries(
            st.binary(),
            st.dictionaries(st.binary(), st.integers(), max_size=3),
            max_size=5,
        )
    )
    def test_dict_items_values_deconstr(self, xs):
        # nested deconstruction with a Value-like object
        source_code = """
def validator(xs: Dict[bytes, Dict[bytes, int]]) -> int:
    sum_values = 0
    for pid, tk_dict in xs.items():
        for tk_name, tk_amount in tk_dict.items():
            sum_values += tk_amount
    return sum_values
"""
        ret = eval_uplc_value(source_code, xs)
        self.assertEqual(
            ret,
            sum(v for pid, d in xs.items() for nam, v in d.items()),
            "for loop deconstruction did not behave as expected",
        )

    def test_no_return_annotation_no_return(self):
        source_code = """
from opshin.prelude import *

def validator(a):
    pass
"""
        res = eval_uplc(source_code, 0)
        self.assertEqual(res, uplc.PlutusConstr(0, []))

    @unittest.expectedFailure
    def test_opt_unsafe_cast(self):
        # test that unsafe casts are not optimized away
        source_code = """
from opshin.prelude import *
def validator(x: Token) -> bool:
    b: Anything = x
    a: int = b
    return True
        """
        ret = eval_uplc(source_code, Unit())

    def test_constant_folding(self):
        source_code = """
from opshin.prelude import *

def validator(_: None) -> bytes:
    return bytes.fromhex("0011")
"""
        res = eval_uplc_value(
            source_code, Unit(), config=DEFAULT_CONFIG_CONSTANT_FOLDING
        )
        self.assertEqual(res, bytes.fromhex("0011"))

    @unittest.expectedFailure
    def test_constant_folding_disabled(self):
        source_code = """
from opshin.prelude import *

def validator(_: None) -> bytes:
    return bytes.fromhex("0011")
"""
        eval_uplc(source_code, Unit(), constant_folding=False)

    def test_constant_folding_list(self):
        source_code = """
from opshin.prelude import *

def validator(_: None) -> List[int]:
    return list(range(0, 10, 2))
"""
        code = builder._compile(
            source_code, Unit(), config=DEFAULT_CONFIG_CONSTANT_FOLDING
        )
        self.assertIn("(con list<integer> [0, 2, 4, 6, 8])", code.dumps())
        res = builder.uplc_eval(code)
        self.assertEqual(
            res, uplc.PlutusList([uplc.PlutusInteger(i) for i in range(0, 10, 2)])
        )

    def test_constant_folding_dict(self):
        source_code = """
from opshin.prelude import *

def validator(_: None) -> Dict[str, bool]:
    return {"s": True, "m": False}
"""
        code = builder._compile(
            source_code, Unit(), config=DEFAULT_CONFIG_CONSTANT_FOLDING
        )
        self.assertIn(
            "(con list<pair<data, data>> [[#4173, #01], [#416d, #00]]))", code.dumps()
        )
        res = uplc_eval(code)
        self.assertEqual(
            res,
            uplc.PlutusMap(
                {
                    uplc.PlutusByteString("s".encode()): uplc.PlutusInteger(1),
                    uplc.PlutusByteString("m".encode()): uplc.PlutusInteger(0),
                }
            ),
        )

    def test_constant_folding_complex(self):
        source_code = """
from opshin.prelude import *

def validator(_: None) -> Dict[str, List[Dict[bytes, int]]]:
    return {"s": [{b"": 0}, {b"0": 1}]}
"""
        res = eval_uplc(source_code, Unit(), config=DEFAULT_CONFIG_CONSTANT_FOLDING)
        self.assertEqual(
            res,
            uplc.PlutusMap(
                {
                    uplc.PlutusByteString("s".encode()): uplc.PlutusList(
                        [
                            uplc.PlutusMap(
                                {uplc.PlutusByteString(b""): uplc.PlutusInteger(0)}
                            ),
                            uplc.PlutusMap(
                                {uplc.PlutusByteString(b"0"): uplc.PlutusInteger(1)}
                            ),
                        ]
                    ),
                }
            ),
        )

    def test_constant_folding_plutusdata(self):
        source_code = """
from opshin.prelude import *

def validator(_: None) -> PubKeyCredential:
    return PubKeyCredential(bytes.fromhex("0011"))
"""
        code = builder._compile(
            source_code, Unit(), config=DEFAULT_CONFIG_CONSTANT_FOLDING
        )
        self.assertIn("(con data #d8799f420011ff)", code.dumps())
        res = uplc_eval(code)
        self.assertEqual(
            res,
            uplc.PlutusConstr(
                constructor=0, fields=[uplc.PlutusByteString(value=b"\x00\x11")]
            ),
        )

    def test_constant_folding_user_def(self):
        source_code = """
def fib(i: int) -> int:
    return i if i < 2 else fib(i-1) + fib(i-2)

def validator(_: None) -> int:
    return fib(10)
"""
        code = builder._compile(
            source_code, Unit(), config=DEFAULT_CONFIG_CONSTANT_FOLDING
        )
        self.assertIn("(con integer 55)", code.dumps())
        res = uplc_eval(code)
        self.assertEqual(
            res.value,
            55,
        )

    @unittest.expectedFailure
    def test_constant_folding_ifelse(self):
        source_code = """
def validator(_: None) -> int:
    if False:
        a = 10
    return a
"""
        eval_uplc(source_code, Unit(), config=DEFAULT_CONFIG_CONSTANT_FOLDING)

    @unittest.expectedFailure
    def test_constant_folding_for(self):
        source_code = """
def validator(x: List[int]) -> int:
    for i in x:
        a = 10
    return a
"""
        eval_uplc(source_code, [], config=DEFAULT_CONFIG_CONSTANT_FOLDING)

    @unittest.expectedFailure
    def test_constant_folding_for_target(self):
        source_code = """
def validator(x: List[int]) -> int:
    for i in x:
        a = 10
    return i
"""
        eval_uplc(source_code, [], config=DEFAULT_CONFIG_CONSTANT_FOLDING)

    @unittest.expectedFailure
    def test_constant_folding_while(self):
        source_code = """
def validator(_: None) -> int:
    while False:
        a = 10
    return a
"""
        eval_uplc(source_code, Unit(), config=DEFAULT_CONFIG_CONSTANT_FOLDING)

    @unittest.skip("Fine from a guarantee perspective, but needs better inspection")
    def test_constant_folding_guaranteed_branch(self):
        source_code = """
def validator(_: None) -> int:
    if False:
        a = 20
        b = 2 * a
    else:
        b = 2
    return b
"""
        code = builder._compile(
            source_code, Unit(), config=DEFAULT_CONFIG_CONSTANT_FOLDING
        )
        self.assertIn("(con integer 40)", code.dumps())

    @unittest.skip("Fine from a guarantee perspective, but needs better inspection")
    def test_constant_folding_scoping(self):
        source_code = """
a = 4
def validator(_: None) -> int:
    a = 2
    b = 5 * a
    return b
"""
        code = builder._compile(
            source_code, Unit(), config=DEFAULT_CONFIG_CONSTANT_FOLDING
        )
        self.assertIn("(con integer 10)", code.dumps())

    @unittest.skip("Fine from a guarantee perspective, but needs better inspection")
    def test_constant_folding_no_scoping(self):
        source_code = """
def validator(_: None) -> int:
    a = 4
    a = 2
    b = 5 * a
    return b
"""
        code = builder._compile(
            source_code, Unit(), config=DEFAULT_CONFIG_CONSTANT_FOLDING
        )
        self.assertIn("(con integer 10)", code.dumps())

    def test_constant_folding_repeated_assign(self):
        source_code = """
def validator(i: int) -> int:
    a = 4
    for k in range(i):
        a = 2
    return a
"""
        code = builder._compile(source_code, config=DEFAULT_CONFIG_CONSTANT_FOLDING)
        res = uplc_eval(uplc.Apply(code, uplc.PlutusInteger(0)))
        self.assertEqual(res.value, 4)
        res = uplc_eval(uplc.Apply(code, uplc.PlutusInteger(1)))
        self.assertEqual(res.value, 2)

    def test_constant_folding_math(self):
        source_code = """
from opshin.prelude import *

def validator(_: None) -> int:
    return 2 ** 10
"""
        code = builder._compile(source_code, config=DEFAULT_CONFIG_CONSTANT_FOLDING)
        code_src = code.dumps()
        self.assertIn(f"(con integer {2**10})", code_src)

    def test_reassign_builtin(self):
        source_code = """
b = int
def validator(_: None) -> int:
    def int(a) -> b:
        return 2
    return int(5)
"""
        res = eval_uplc_value(source_code, Unit())
        self.assertEqual(res, 2)

    @unittest.expectedFailure
    def test_reassign_builtin_invalid_type(self):
        source_code = """
def validator(_: None) -> int:
    def int(a) -> int:
        return 2
    return int(5)
"""
        builder._compile(source_code)

    def test_constant_folding_ignore_reassignment(self):
        source_code = """
b = int
def validator(_: None) -> int:
    def int(a) -> b:
        return 2
    return int(5)
"""
        res = eval_uplc_value(
            source_code, Unit(), config=DEFAULT_CONFIG_CONSTANT_FOLDING
        )
        self.assertEqual(res, 2)

    def test_constant_folding_no_print_eval(self):
        source_code = """
from opshin.prelude import *

def validator(_: None) -> None:
    return print("hello")
"""
        code = builder._compile(source_code, config=DEFAULT_CONFIG_CONSTANT_FOLDING)
        code_src = code.dumps()
        self.assertIn(f'(con string "hello")', code_src)

    def test_inner_outer_state_functions(self):
        source_code = """
a = 2
def b() -> int:
    return a

def validator(_: None) -> int:
    a = 3
    return b()
        """
        res = eval_uplc_value(source_code, Unit())
        self.assertEqual(res, 2)

    def test_inner_outer_state_functions_nonglobal(self):
        source_code = """

def validator(_: None) -> int:
    a = 2
    def b() -> int:
        return a
    def c() -> int:
        a = 3
        return b()
    return c()
"""
        res = eval_uplc_value(source_code, Unit())
        self.assertEqual(res, 2)

    def test_outer_state_change_functions(self):
        source_code = """
a = 2
def b() -> int:
    return a
a = 3

def validator(_: None) -> int:
    return b()
"""
        res = eval_uplc_value(source_code, Unit())
        self.assertEqual(res, 3)

    @unittest.expectedFailure
    def test_failing_annotated_type(self):
        source_code = """
def c():
    a = 2
    def b() -> int:
        return a
    return b

def validator(_: None):
    a = 3
    return c()
"""
        eval_uplc(source_code, Unit())

    @unittest.expectedFailure
    def test_access_enclosing_variable_before_def(self):
        # note this is a runtime error, just like it would be in python!
        source_code = """
a = "1"
def validator(_: None) -> None:
   def d() -> str:
       return a
   print(d())
   a = "2"
"""
        eval_uplc(source_code, Unit())

    @unittest.expectedFailure
    def test_access_local_variable_before_assignment(self):
        # note this is a runtime error, just like it would be in python!
        source_code = """
a = "1"
def validator(_: None) -> None:
   print(a)
   a = "2"
"""
        eval_uplc(source_code, Unit())

    def test_warn_bytestring(self):
        source_code = """
b = b"0011ff"
def validator(_: None) -> None:
    pass
"""
        eval_uplc(source_code, Unit())

    @parameterized.expand(
        [
            (
                "d8799fd8799f9fd8799fd8799fd8799f582055d353acacaab6460b37ed0f0e3a1a0aabf056df4a7fa1e265d21149ccacc527ff01ffd8799fd8799fd87a9f581cdbe769758f26efb21f008dc097bb194cffc622acc37fcefc5372eee3ffd87a80ffa140a1401a00989680d87a9f5820dfab81872ce2bbe6ee5af9bbfee4047f91c1f57db5e30da727d5fef1e7f02f4dffd87a80ffffff809fd8799fd8799fd8799f581cdc315c289fee4484eda07038393f21dc4e572aff292d7926018725c2ffd87a80ffa140a14000d87980d87a80ffffa140a14000a140a1400080a0d8799fd8799fd87980d87a80ffd8799fd87b80d87a80ffff80a1d87a9fd8799fd8799f582055d353acacaab6460b37ed0f0e3a1a0aabf056df4a7fa1e265d21149ccacc527ff01ffffd87980a15820dfab81872ce2bbe6ee5af9bbfee4047f91c1f57db5e30da727d5fef1e7f02f4dd8799f581cdc315c289fee4484eda07038393f21dc4e572aff292d7926018725c2ffd8799f5820746957f0eb57f2b11119684e611a98f373afea93473fefbb7632d579af2f6259ffffd87a9fd8799fd8799f582055d353acacaab6460b37ed0f0e3a1a0aabf056df4a7fa1e265d21149ccacc527ff01ffffff"
            ),
            (
                "d8799fd8799f9fd8799fd8799fd8799f582055d353acacaab6460b37ed0f0e3a1a0aabf056df4a7fa1e265d21149ccacc527ff01ffd8799fd8799fd87a9f581cdbe769758f26efb21f008dc097bb194cffc622acc37fcefc5372eee3ffd87a80ffa140a1401a00989680d87a9f5820dfab81872ce2bbe6ee5af9bbfee4047f91c1f57db5e30da727d5fef1e7f02f4dffd87a80ffffff809fd8799fd8799fd8799f581cdc315c289fee4484eda07038393f21dc4e572aff292d7926018725c2ffd87a80ffa140a14000d87980d87a80ffffa140a14000a140a1400080a0d8799fd8799fd87a9f1b000001836ac117d8ffd87a80ffd8799fd87b80d87a80ffff80a1d87a9fd8799fd8799f582055d353acacaab6460b37ed0f0e3a1a0aabf056df4a7fa1e265d21149ccacc527ff01ffffd87980a15820dfab81872ce2bbe6ee5af9bbfee4047f91c1f57db5e30da727d5fef1e7f02f4dd8799f581cdc315c289fee4484eda07038393f21dc4e572aff292d7926018725c2ffd8799f5820797a1e1720b63621c6b185088184cb8e23af6e46b55bd83e7a91024c823a6c2affffd87a9fd8799fd8799f582055d353acacaab6460b37ed0f0e3a1a0aabf056df4a7fa1e265d21149ccacc527ff01ffffff"
            ),
            (
                "d8799fd8799f9fd8799fd8799fd8799f582055d353acacaab6460b37ed0f0e3a1a0aabf056df4a7fa1e265d21149ccacc527ff01ffd8799fd8799fd87a9f581cdbe769758f26efb21f008dc097bb194cffc622acc37fcefc5372eee3ffd87a80ffa140a1401a00989680d87a9f5820dfab81872ce2bbe6ee5af9bbfee4047f91c1f57db5e30da727d5fef1e7f02f4dffd87a80ffffff809fd8799fd8799fd8799f581cdc315c289fee4484eda07038393f21dc4e572aff292d7926018725c2ffd87a80ffa140a14000d87980d87a80ffd8799fd8799fd8799f581cdc315c289fee4484eda07038393f21dc4e572aff292d7926018725c2ffd87a80ffa140a1401a000f4240d87980d87a80ffffa140a14000a140a1400080a0d8799fd8799fd87a9f1b000001836ac117d8ffd87a80ffd8799fd87b80d87a80ffff9f581cdc315c289fee4484eda07038393f21dc4e572aff292d7926018725c2ffa1d87a9fd8799fd8799f582055d353acacaab6460b37ed0f0e3a1a0aabf056df4a7fa1e265d21149ccacc527ff01ffffd87980a15820dfab81872ce2bbe6ee5af9bbfee4047f91c1f57db5e30da727d5fef1e7f02f4dd8799f581cdc315c289fee4484eda07038393f21dc4e572aff292d7926018725c2ffd8799f5820c17c32f6433ae22c2acaebfb796bbfaee3993ff7ebb58a2bac6b4a3bdd2f6d28ffffd87a9fd8799fd8799f582055d353acacaab6460b37ed0f0e3a1a0aabf056df4a7fa1e265d21149ccacc527ff01ffffff"
            ),
        ]
    )
    def test_script_context_str_format(self, p: str):
        context = ScriptContext.from_cbor(bytes.fromhex(p))
        source_code = """
from opshin.prelude import *

def validator(c: ScriptContext) -> str:
    return f"{c}"
        """
        res = eval_uplc_value(source_code, context)
        # should not raise
        eval(res)

    @hypothesis.given(st.binary(), st.binary())
    def test_uplc_builtin(self, x, y):
        source_code = """
from opshin.std.builtins import *
def validator(x: bytes, y: bytes) -> bytes:
    return append_byte_string(x, y)
"""
        res = eval_uplc_value(source_code, x, y)
        self.assertEqual(res, x + y)

    def test_trace_order(self):
        # TODO can become a proper test once uplc is upgraded to >=1.0.0
        source_code = """
from opshin.std.builtins import *
def validator() -> None:
    print("test")
    print("hi")
    print("there")
    return None
"""
        eval_uplc(source_code, PlutusData())

    def test_print_empty(self):
        # TODO can become a proper test once uplc is upgraded to >=1.0.0
        source_code = """
from opshin.std.builtins import *
def validator() -> None:
    print()
    print()
    print()
    print()
    print()
    print()
    return None
"""
        eval_uplc(source_code, PlutusData())

    @hypothesis.given(st.integers())
    def test_cast_bool_ite(self, x):
        source_code = """
def validator(x: int) -> bool:
    if x:
        res = True
    else:
        res = False
    return res
"""
        res = eval_uplc_value(source_code, x)
        self.assertEqual(res, bool(x))

    @hypothesis.given(st.integers())
    def test_cast_bool_ite_expr(self, x):
        source_code = """
def validator(x: int) -> bool:
    return True if x else False
"""
        res = eval_uplc_value(source_code, x)
        self.assertEqual(bool(res), bool(x))

    @hypothesis.given(st.integers())
    def test_cast_bool_while(self, x):
        source_code = """
def validator(x: int) -> bool:
    res = False
    while x:
        res = True
        x = 0
    return res
"""
        res = eval_uplc_value(source_code, x)
        self.assertEqual(bool(res), bool(x))

    @hypothesis.given(st.integers())
    def test_cast_bool_boolops(self, x):
        source_code = """
def validator(x: int) -> bool:
    return x and x or (x or x)
"""
        res = eval_uplc_value(source_code, x)
        self.assertEqual(bool(res), bool(x and x or (x or x)))

    @hypothesis.given(st.integers())
    def test_cast_bool_ite(self, x):
        source_code = """
def validator(x: int) -> None:
    assert x
"""
        try:
            eval_uplc(source_code, x)
            res = True
        except Exception:
            res = False
        self.assertEqual(res, bool(x))

    @hypothesis.given(a_or_b)
    def test_isinstance_cast_if(self, x):
        source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: int
    bar: int

def validator(x: Union[A, B]) -> int:
    if isinstance(x, A):
        k = x.foo
    elif isinstance(x, B):
        k = x.bar
    return k
"""
        res = eval_uplc_value(source_code, x)
        self.assertEqual(res, x.foo if isinstance(x, A) else x.bar)

    @hypothesis.given(a_or_b, a_or_b)
    def test_complex_isinstance_cast_if(self, x, y):
        source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: int
    bar: int

def validator(x: Union[A, B], y: Union[A, B]) -> int:
    if isinstance(x, A) and isinstance(y, B):
        k = x.foo + y.bar
    elif isinstance(x, A) and isinstance(y, A):
        k = x.foo + y.foo
    elif isinstance(x, B) and isinstance(y, A):
        k = x.bar + y.foo
    elif isinstance(x, B) and isinstance(y, B):
        k = x.bar + y.bar
    return k
"""
        res = eval_uplc_value(source_code, x, y)
        self.assertEqual(
            res,
            (x.foo if isinstance(x, A) else x.bar)
            + (y.foo if isinstance(y, A) else y.bar),
        )

    @hypothesis.given(a_or_b)
    def test_isinstance_cast_ifexpr(self, x):
        source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: int
    bar: int

def validator(x: Union[A, B]) -> int:
    k = x.foo if isinstance(x, A) else x.bar if isinstance(x, B) else 0
    return k
"""
        res = eval_uplc_value(source_code, x)
        self.assertEqual(res, x.foo if isinstance(x, A) else x.bar)

    @hypothesis.given(a_or_b)
    def test_isinstance_cast_while(self, x):
        source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: int
    bar: int

def validator(x: Union[A, B]) -> int:
    foo = 0
    while isinstance(x, B) and foo != 1:
        foo = x.bar
        foo = 1
    return foo
"""
        res = eval_uplc_value(source_code, x)
        self.assertEqual(res, 1 if isinstance(x, B) else 0)

    @hypothesis.given(a_or_b)
    def test_isinstance_cast_random(self, x):
        source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: int
    bar: int

def validator(x: Union[A, B]) -> bool:
    return isinstance(x, A)
"""
        res = eval_uplc_value(source_code, x)
        self.assertEqual(res, isinstance(x, A))

    @hypothesis.given(a_or_b, st.integers())
    @hypothesis.example(A(0), 0)
    def test_isinstance_cast_shortcut_and(self, x, y):
        source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: int
    bar: int

def validator(x: Union[A, B], y: int) -> bool:
    return isinstance(x, A) and x.foo == y or isinstance(x, B) and x.bar == y
"""
        res = eval_uplc_value(source_code, x, y)
        self.assertEqual(
            res, isinstance(x, A) and x.foo == y or isinstance(x, B) and x.bar == y
        )

    @hypothesis.given(a_or_b)
    def test_isinstance_cast_assert(self, x):
        source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: int
    bar: int

def validator(x: Union[A, B]) -> int:
    assert isinstance(x, B), "Wrong type"
    return x.bar
"""
        try:
            res = eval_uplc_value(source_code, x)
        except:
            res = None
        self.assertEqual(res, x.bar if isinstance(x, B) else None)

    @unittest.expectedFailure
    def test_isinstance_cast_assert_if(self):
        source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: int
    bar: int

def validator(x: Union[A, B]) -> int:
    if True:
        assert isinstance(x, B), "Wrong type"
    return x.bar
"""
        builder._compile(source_code)

    @hypothesis.given(a_or_b)
    def test_isinstance_cast_complex_or(self, x):
        source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: int
    bar: int
    
@dataclass()
class C(PlutusData):
    CONSTR_ID = 2
    foo: int

def validator(x: Union[A, B, C]) -> int:
    if isinstance(x, A) or isinstance(x, C):
        res = x.foo
    else:
        res = 100
    return res
"""
        res = eval_uplc_value(source_code, x)
        self.assertEqual(res, x.foo if isinstance(x, A) else 100)

    @unittest.expectedFailure
    def test_isinstance_cast_complex_or_sameconstr(self):
        source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: int
    bar: int

@dataclass()
class C(PlutusData):
    CONSTR_ID = 0
    foo: int

def validator(x: Union[A, B]) -> int:
    if isinstance(x, A) or isinstance(x, C):
        res = x.foo
    else:
        res = 100
    return res
"""
        builder._compile(source_code)
        print("Union of same constructor id was allowed, should be disallowed")

    @hypothesis.given(a_or_b)
    def test_isinstance_cast_complex_not(self, x):
        source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: int
    bar: int
    
def validator(x: Union[A, B]) -> int:
    if not isinstance(x, B):
        res = x.foo
    else:
        res = 100
    return res
"""
        res = eval_uplc_value(source_code, x)
        self.assertEqual(res, x.foo if not isinstance(x, B) else 100)

    @hypothesis.given(a_or_b)
    def test_isinstance_cast_complex_ifelse(self, x):
        source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: int
    bar: int

def validator(x: Union[A, B]) -> int:
    if isinstance(x, A):
        res = x.foo
    else:
        res = x.bar
    return res
"""
        res = eval_uplc_value(source_code, x)
        self.assertEqual(res, x.foo if isinstance(x, A) else x.bar)

    @unittest.expectedFailure
    def test_isinstance_cast_complex_or_else(self):
        source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: int
    bar: int

def validator(x: Union[A, B]) -> int:
    foo = 0
    if isinstance(x, B) or foo == 0:
        foo = x.bar
    else:
        foo = x.foo
    return foo
"""
        builder._compile(source_code)

    @unittest.expectedFailure
    def test_isinstance_cast_complex_and_else(self):
        source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: int
    bar: int

def validator(x: Union[A, B]) -> int:
    foo = 0
    if isinstance(x, B) and foo == 0:
        foo = x.bar
    else:
        foo = x.foo
    return foo
"""
        builder._compile(source_code)

    def test_isinstance_cast_complex_and(self):
        source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: int
    bar: int

def validator(x: Union[A, B]) -> int:
    foo = 0
    if isinstance(x, B) and foo == 0:
        foo = x.bar
    return foo
"""
        builder._compile(source_code)

    @hypothesis.given(a_or_b, st.integers())
    def test_isinstance_cast_shortcut_or(self, x, y):
        source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: int
    bar: int

def validator(x: Union[A, B], y: int) -> bool:
    return (isinstance(x, A) or x.bar == y) and (isinstance(x, B) or x.foo == y)
"""
        res = eval_uplc_value(source_code, x, y)
        self.assertEqual(
            res, (isinstance(x, A) or x.bar == y) and (isinstance(x, B) or x.foo == y)
        )

    @hypothesis.given(a_or_b)
    def test_uniontype_if(self, x):
        source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: int
    bar: int

def validator(x: Union[A, B]) -> Union[A, B]:
    if isinstance(x, A):
        k = B(x.foo, 1)
    else:
        k = A(x.bar)
    return k
"""
        res = uplc.plutus_cbor_dumps(eval_uplc(source_code, x))
        self.assertEqual(res, (B(x.foo, 1) if isinstance(x, A) else A(x.bar)).to_cbor())

    @unittest.expectedFailure
    def test_if_no_retype_no_plutusdata(self):
        source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: int
    bar: int

def validator(x: Union[A, B]):
    if isinstance(x, A):
        k = B(x.foo, 1)
    else:
        k = "hello"
    return k
"""
        builder._compile(source_code)

    @unittest.expectedFailure
    def test_while_no_retype_no_plutusdata(self):
        source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: int
    bar: int

def validator(x: Union[A, B]):
    while isinstance(x, A):
        k = B(x.foo, 1)
    else:
        k = "hello"
    return k
"""
        builder._compile(source_code)

    @unittest.expectedFailure
    @hypothesis.given(a_or_b)
    def test_retype_while(self, x):
        source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: int
    bar: int

def validator(x: Union[A, B]) -> int:
    while isinstance(x, A):
        x = B(x.foo, 1)
    return x.foobar
"""
        res = eval_uplc_value(source_code, x)
        self.assertEqual(res, x.foo if isinstance(x, A) else x.foobar)

    @unittest.expectedFailure
    def test_retype_if_branch_correct(self):
        source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: int
    bar: int

def validator(x: Union[A, B]) -> int:
    if False:
        x = B(0, 1)
    return x.foobar
"""
        builder._compile(source_code)

    @unittest.expectedFailure
    def test_retype_while_branch_correct(self):
        source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: int
    bar: int

def validator(x: Union[A, B]) -> int:
    while False:
        x = B(0, 1)
    return x.foobar
"""
        builder._compile(source_code)

    @unittest.expectedFailure
    def test_retype_while_wrong_after_iter(self):
        source_code = """
def validator(x: int) -> bytes:
    while True:
        x += 1
        x = b''
    return x
"""
        builder._compile(source_code)

    @unittest.expectedFailure
    def test_retype(self):
        source_code = """
def validator(x: int) -> str:
    x = "hello"
    return x
"""
        res = eval_uplc_value(source_code, 1)
        self.assertEqual(res, b"hello")

    @unittest.expectedFailure
    def test_retype_if_primitives(self):
        source_code = """
def validator(x: int) -> str:
    if True:
        x = "hello"
    else:
        x = "hi"
    return x
"""
        res = eval_uplc_value(source_code, 1)
        self.assertEqual(res, b"hello")

    @unittest.expectedFailure
    def test_in_list(self):
        source_code = """
from opshin.prelude import *

def validator(
    d: Nothing,
    r: Nothing,
    context: ScriptContext,
):
    assert context.purpose in context.tx_info.signatories
"""
        builder._compile(source_code)

    @hypothesis.given(
        st.lists(
            st.tuples(
                st.integers(), st.sampled_from(["<", "<=", "==", ">=", ">", "!="])
            ),
            max_size=10,
            min_size=2,
        )
    )
    @hypothesis.example(
        [
            (0, "<"),
            (0, "<"),
            (0, "<"),
            (0, "<"),
            (0, "<"),
            (0, "<"),
            (0, "<"),
            (0, "<"),
            (0, "<"),
            (0, "<"),
        ],
    )
    def test_comparison_chaining(self, xs):
        param_string = ",".join(f"i{k}: int" for k, _ in enumerate(xs))
        comp_string = "i0"
        eval_string = f"{xs[0][0]}"
        for k, (x, c) in enumerate(xs[1:], start=1):
            comp_string += f" {c} i{k}"
            eval_string += f" {c} {x}"
        source_code = f"""
def validator({param_string}) -> bool:
    return {comp_string}
"""
        res = eval_uplc_value(source_code, *[x[0] for x in xs])
        self.assertEqual(bool(res), eval(eval_string))

    def test_double_import_offset(self):
        source_code = """
from opshin.ledger.api_v2 import *
from opshin.prelude import *

def validator(
    d: Nothing,
    r: Nothing,
    context: ScriptContext,
):
    house_address = Address(
        payment_credential=PubKeyCredential(
            credential_hash=b""
        ),
        staking_credential=SomeStakingCredential(
            staking_credential=StakingHash(
                value=PubKeyCredential(
                    credential_hash=b""
                )
            )
        ),
    )
"""
        # would fail because Address is assigned multiple times and then not constant folded
        # TODO find a better way
        builder._compile(source_code, config=DEFAULT_CONFIG_CONSTANT_FOLDING)

    def test_double_import_direct(self):
        source_code = """
from opshin.prelude import *
from opshin.prelude import *

def validator(
    d: Nothing,
    r: Nothing,
    context: ScriptContext,
):
    house_address = Address(
        payment_credential=PubKeyCredential(
            credential_hash=b""
        ),
        staking_credential=SomeStakingCredential(
            staking_credential=StakingHash(
                value=PubKeyCredential(
                    credential_hash=b""
                )
            )
        ),
    )
"""
        # would fail because Address is assigned multiple times and then not constant folded
        # TODO find a better way
        builder._compile(source_code, config=DEFAULT_CONFIG_CONSTANT_FOLDING)

    def test_double_import_deep(self):
        source_code = """
from opshin.ledger.interval import *
from opshin.prelude import *

def validator(
    d: Nothing,
    r: Nothing,
    context: ScriptContext,
):
    house_address = Address(
        payment_credential=PubKeyCredential(
            credential_hash=b""
        ),
        staking_credential=SomeStakingCredential(
            staking_credential=StakingHash(
                value=PubKeyCredential(
                    credential_hash=b""
                )
            )
        ),
    )
"""
        # would fail because Address is assigned multiple times and then not constant folded
        # TODO find a better way
        builder._compile(source_code, config=DEFAULT_CONFIG_CONSTANT_FOLDING)

    def test_bytearray_alternative(self):
        source_code = """
def validator(
    d: bytearray,
) -> bytes:
    return d
"""
        eval_uplc(source_code, bytearray(b"hello"))

    # TODO enable when pycardano version is fixed s.t. import of ByteString works
    @unittest.expectedFailure
    def test_ByteString_alternative(self):
        source_code = """
def validator(
    d: ByteString,
) -> bytes:
    return d
"""
        eval_uplc(source_code, bytearray(b"hello"))

    @hypothesis.given(
        st.lists(
            st.tuples(st.booleans(), st.sampled_from(["and", "or"])),
            max_size=10,
            min_size=2,
        )
    )
    def test_boolop_chaining(self, xs):
        param_string = ",".join(f"i{k}: bool" for k, _ in enumerate(xs))
        comp_string = "i0"
        eval_string = f"{xs[0][0]}"
        for k, (x, c) in enumerate(xs[1:], start=1):
            comp_string += f" {c} i{k}"
            eval_string += f" {c} {x}"
        source_code = f"""
def validator({param_string}) -> bool:
    return {comp_string}
"""
        res = eval_uplc_value(source_code, *[x[0] for x in xs])
        self.assertEqual(bool(res), eval(eval_string))

    def test_wrapping_contract_apply(self):
        # TODO devise tests for this
        input_file = "examples/smart_contracts/wrapped_token.py"
        contract = builder.build(input_file, config=DEFAULT_CONFIG_FORCE_THREE_PARAMS)
        artifacts = PlutusContract(
            contract,
            datum_type=("datum", prelude.Nothing),
            redeemer_type=("redeemer", prelude.Nothing),
            parameter_types=[
                ("token_policy_id", bytes),
                ("token_name", bytes),
                ("wrapping_factor", int),
            ],
            purpose=[Purpose.spending, Purpose.minting],
        )
        applied = artifacts.apply_parameter(b"", b"")
        assert len(applied.parameter_types) == 1
        assert applied.parameter_types[0][0] == "wrapping_factor"
        assert applied.datum_type == ("datum", prelude.Nothing)

    def test_wrapping_contract_dump_load(self):
        input_file = "examples/smart_contracts/wrapped_token.py"
        contract = builder.build(input_file, config=DEFAULT_CONFIG_FORCE_THREE_PARAMS)
        artifacts = PlutusContract(
            contract,
            datum_type=("datum", prelude.Nothing),
            redeemer_type=("redeemer", prelude.ScriptContext),
            parameter_types=[
                ("token_policy_id", bytes),
                ("token_name", bytes),
                ("wrapping_factor", int),
            ],
            purpose=[Purpose.spending, Purpose.minting],
            description="Wrapped token contract",
            license="MIT",
        )
        target_dir = tempfile.TemporaryDirectory()
        artifacts.dump(target_dir.name)
        loaded = builder.load(target_dir.name)
        assert len(loaded.parameter_types) == len(artifacts.parameter_types)
        assert loaded.datum_type[1].__name__ == artifacts.datum_type[1].__name__
        assert loaded.datum_type[0] == artifacts.datum_type[0]
        assert loaded.datum_type[1].CONSTR_ID == artifacts.datum_type[1].CONSTR_ID
        assert loaded.redeemer_type[1].__name__ == artifacts.redeemer_type[1].__name__
        assert loaded.redeemer_type[0] == artifacts.redeemer_type[0]
        assert loaded.redeemer_type[1].CONSTR_ID == artifacts.redeemer_type[1].CONSTR_ID
        assert loaded.purpose == artifacts.purpose
        assert loaded.description == artifacts.description
        assert loaded.license == artifacts.license
        assert loaded.title == artifacts.title
        assert loaded.version == artifacts.version

    def test_bridge(self):
        @wraps_builtin
        def add_integer(x: int, y: int) -> int:
            return x + y

        assert add_integer(1, 2) == 3

        @wraps_builtin
        def append_byte_string(x: bytes, y: bytes) -> bytes:
            return x + y

        assert append_byte_string(b"hello", b"world") == b"helloworld"

    @unittest.expectedFailure
    def test_forbidden_overwrite(self):
        source_code = """
def validator(
    d: int
):
    PlutusData = d
    return d
"""
        builder._compile(source_code)

    @parameterized.expand(ALL_EXAMPLES)
    def test_compilation_deterministic_local(self, input_file):
        with open(input_file) as fp:
            source_code = fp.read()
        code = builder._compile(source_code)
        for i in range(10):
            code_2 = builder._compile(source_code)
            self.assertEqual(code.dumps(), code_2.dumps())

    @parameterized.expand(ALL_EXAMPLES)
    def test_compilation_deterministic_external(self, input_file):
        code = subprocess.run(
            [
                sys.executable,
                "-m",
                "opshin",
                "compile",
                "any",
                input_file,
            ],
            capture_output=True,
        )
        for i in range(10):
            code_2 = subprocess.run(
                [
                    sys.executable,
                    "-m",
                    "opshin",
                    "compile",
                    "any",
                    input_file,
                ],
                capture_output=True,
            )
            self.assertEqual(code.stdout, code_2.stdout)

    @unittest.expectedFailure
    def test_return_illegal(self):
        # this is now an illegal retyping because read variables dont match
        source_code = """
return 1
def validator(_: None) -> int:
    return 0
        """
        builder._compile(source_code)

    def test_return_in_loop(self):
        source_code = """
def validator(_: None) -> int:
    i = 0
    while i < 10:
        i += 1
        if i == 5:
          return i
    return 0
        """
        res = eval_uplc_value(source_code, Unit())
        self.assertEqual(res, 5, "Invalid return break")

    def test_return_in_for(self):
        source_code = """
def validator(_: None) -> int:
    i = 0
    for i in range(10):
        i += 1
        if i == 5:
          return i
    return 0
        """
        res = eval_uplc_value(source_code, Unit())
        self.assertEqual(res, 5, "Invalid return break")

    def test_return_in_if(self):
        source_code = """
def validator(_: None) -> int:
    i = 0
    if i == 1:
        return 0
    else:
        return 1
        """
        res = eval_uplc_value(source_code, Unit())
        self.assertEqual(res, 1, "Invalid return")

    @unittest.expectedFailure
    def test_return_in_if_same_type(self):
        source_code = """
def validator(_: None) -> str:
    i = 0
    if i == 1:
        return "a"
    else:
        return 1
        """
        builder._compile(source_code)

    def test_isinstance_cast_if2(self):
        source_code = """
from dataclasses import dataclass
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    foobar: int
    bar: int

def validator(_: None) -> Union[A, B]:
    x = 0
    if x == 1:
        return A(1)
    else:
        return B(2, 1)
"""
        res = eval_uplc(source_code, Unit())
        self.assertEqual(
            res,
            uplc.PlutusConstr(1, [uplc.PlutusInteger(2), uplc.PlutusInteger(1)]),
            "Invalid return",
        )

    @unittest.expectedFailure
    def test_return_in_if_missing_return(self):
        source_code = """
def validator(_: None) -> str:
    i = 0
    if i == 1:
        return "a"
    else:
        pass
        """
        builder._compile(source_code)

    def test_different_return_types_anything(self):
        source_code = """
from opshin.prelude import *

def validator(a: int) -> Anything:
    if a > 0:
        return b""
    else:
        return 0
"""
        res = eval_uplc(source_code, 1)
        self.assertEqual(res, uplc.PlutusByteString(b""))
        res = eval_uplc(source_code, -1)
        self.assertEqual(res, uplc.PlutusInteger(0))

    @unittest.expectedFailure
    def test_different_return_types_while_loop(self):
        source_code = """
def validator(a: int) -> str:
    while a > 0:
        return b""
    return 0
"""
        builder._compile(source_code)

    @unittest.expectedFailure
    def test_different_return_types_for_loop(self):
        source_code = """
def validator(a: int) -> str:
    for i in range(a):
        return b""
    return 0
"""
        builder._compile(source_code)

    def test_return_else_loop_while(self):
        source_code = """
def validator(a: int) -> int:
    while a > 0:
        a -= 1
    else:
        return 0
"""
        res = eval_uplc_value(source_code, 1)
        self.assertEqual(res, 0, "Invalid return")

    def test_return_else_loop_for(self):
        source_code = """
def validator(a: int) -> int:
    for _ in range(a):
        a -= 1
    else:
        return 0
"""
        res = eval_uplc_value(source_code, 1)
        self.assertEqual(res, 0, "Invalid return")

    def test_empty_list_int(self):
        source_code = """
def validator(_: None) -> List[int]:
    a: List[int] = []
    return a + [1]
"""
        res = eval_uplc_value(source_code, Unit())
        self.assertEqual(res, [uplc.PlutusInteger(1)])

    def test_empty_list_data(self):
        source_code = """
from opshin.prelude import *

def validator(_: None) -> List[Token]:
    a: List[Token] = []
    return a + [Token(b"", b"")]
"""
        res = eval_uplc_value(source_code, Unit())
        self.assertEqual(
            res,
            [
                uplc.PlutusConstr(
                    0, [uplc.PlutusByteString(b""), uplc.PlutusByteString(b"")]
                )
            ],
        )

    def test_empty_list_int_constant_folding(self):
        source_code = """
def validator(_: None) -> List[int]:
    a: List[int] = []
    return a
"""
        res = eval_uplc_value(
            source_code, Unit(), config=DEFAULT_CONFIG_CONSTANT_FOLDING
        )
        self.assertEqual(res, [])

    def test_empty_dict_int_int(self):
        source_code = """
def validator(_: None) -> Dict[int, int]:
    a: Dict[int, int] = {}
    return a
"""
        res = eval_uplc_value(source_code, Unit())
        self.assertEqual(res, {})

    def test_empty_dict_int_int_constant_folding(self):
        source_code = """
def validator(_: None) -> Dict[int, int]:
    a: Dict[int, int] = {}
    return a
"""
        res = eval_uplc_value(
            source_code, Unit(), config=DEFAULT_CONFIG_CONSTANT_FOLDING
        )
        self.assertEqual(res, {})

    def test_empty_dict_displaced_constant_folding(self):
        source_code = """
from typing import Dict, List, Union

VAR: Dict[bytes, int] = {}

def validator(b: Dict[int, Dict[bytes, int]]) -> Dict[bytes, int]:
    a = b.get(0, VAR)
    return a
        """
        res = eval_uplc_value(
            source_code, {1: {b"": 0}}, config=DEFAULT_CONFIG_CONSTANT_FOLDING
        )
        self.assertEqual(res, {})

    def test_union_subset_call(self):
        source_code = """
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData
from dataclasses import dataclass

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    bar: int

@dataclass()
class C(PlutusData):
    CONSTR_ID = 2
    foobar: int

def fun(x: Union[A, B, C]) -> int:
    return 0


def validator(x: Union[A, B]) -> int:
    return fun(x)
        """
        builder._compile(source_code)

    @unittest.expectedFailure
    def test_union_superset_call(self):
        source_code = """
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData
from dataclasses import dataclass

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    bar: int

@dataclass()
class C(PlutusData):
    CONSTR_ID = 2
    foobar: int

def fun(x: Union[A, B]) -> int:
    return 0


def validator(x: Union[A, B, C]) -> int:
    return fun(x)
        """
        builder._compile(source_code)

    @unittest.expectedFailure
    def test_merge_function_same_capture_different_type(self):
        source_code = """
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData
from dataclasses import dataclass

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    bar: int

def validator(x: bool) -> int:
    if x:
        y = A(0)
        def foo() -> int:
            return y.foo
    else:
        y = B(0)
        def foo() -> int:
            return y.bar
    y = A(0)
    return foo()
        """
        builder._compile(source_code)

    def test_merge_function_same_capture_same_type(self):
        source_code = """
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData
from dataclasses import dataclass

@dataclass()
class A(PlutusData):
    CONSTR_ID = 0
    foo: int

@dataclass()
class B(PlutusData):
    CONSTR_ID = 1
    bar: int

def validator(x: bool) -> int:
    if x:
        y = A(0)
        def foo() -> int:
            print(2)
            return y.foo
    else:
        y = A(0) if x else B(0)
        def foo() -> int:
            print(y)
            return 2
    y = A(0)
    return foo()
        """
        res_true = eval_uplc_value(source_code, 1)
        res_false = eval_uplc_value(source_code, 0)
        self.assertEqual(res_true, 0)
        self.assertEqual(res_false, 2)

    def test_merge_print(self):
        source_code = """
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData
from dataclasses import dataclass

def validator(x: bool) -> None:
    if x:
        a = print
    else:
        b = print
        a = b
    return a(x)
        """
        res_true = eval_uplc(source_code, 1)
        res_false = eval_uplc(source_code, 0)

    def test_print_reassign(self):
        source_code = """
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData
from dataclasses import dataclass

def validator(x: bool) -> None:
    a = print
    return a(x)
        """
        res_true = eval_uplc(source_code, 1)
        res_false = eval_uplc(source_code, 0)

    def test_str_constr_reassign(self):
        source_code = """
from typing import Dict, List, Union
from pycardano import Datum as Anything, PlutusData
from dataclasses import dataclass

def validator(x: bool) -> str:
    a = str
    return a(x)
        """
        res_true = eval_uplc_value(source_code, 1)
        res_false = eval_uplc_value(source_code, 0)

    @unittest.expectedFailure
    def test_class_attribute_access(self):
        source_code = """
from dataclasses import dataclass
from pycardano import Datum as Anything, PlutusData
from typing import Dict, List, Union

@dataclass
class A(PlutusData):
    CONSTR_ID = 0
    a: int
    b: bytes
    d: List[int]

def validator(_: None) -> int:
    return A.CONSTR_ID
    """
        builder._compile(source_code)

    def test_constr_id_access(self):
        source_code = """
from dataclasses import dataclass
from pycardano import Datum as Anything, PlutusData
from typing import Dict, List, Union

@dataclass
class A(PlutusData):
    CONSTR_ID = 15
    a: int
    b: bytes
    d: List[int]

def validator(_: None) -> int:
    return A(0, b"", [1,2]).CONSTR_ID
    """
        res = eval_uplc_value(source_code, Unit())

        self.assertEqual(15, res, "Invalid constr id")

    def test_id_map_equals_pycardano(self):
        @dataclass
        class A(PlutusData):
            CONSTR_ID = 0
            a: int
            b: bytes
            d: List[int]

        @dataclass
        class C(PlutusData):
            z: Anything

        @dataclass
        class B(PlutusData):
            a: int
            c: A
            d: Dict[bytes, C]
            e: Union[A, C]

        source_code = """
from dataclasses import dataclass
from pycardano import Datum as Anything, PlutusData
from typing import Dict, List, Union

@dataclass
class Nothing(PlutusData):
    CONSTR_ID = 0
    

@dataclass
class A(PlutusData):
    CONSTR_ID = 0
    a: int
    b: bytes
    d: List[int]

@dataclass
class C(PlutusData):
    z: Anything

@dataclass
class B(PlutusData):
    a: int
    c: A
    d: Dict[bytes, C]
    e: Union[A, C]

def validator(_: None) -> int:
    return B(1, A(1, b"", [1, 2]), {b"": C(Nothing())}, C(Nothing())).CONSTR_ID
    """
        res = eval_uplc_value(source_code, Unit())

        self.assertEqual(
            B.CONSTR_ID, res, "Invalid constr id generation (does not match pycardano)"
        )

    def test_empty_return(self):
        source_code = """
def validator(_: None) -> None:
    return
"""
        res = eval_uplc(source_code, Unit())
        self.assertEqual(res, uplc.PlutusConstr(0, []), "Invalid return")



from copy import copy

from ..typed_ast import *
from ..util import (
    CompilingNodeTransformer,
)
from ..fun_impls import PythonBuiltIn, PythonBuiltInTypes

"""
Inject initialising the builtin functions
"""


class RewriteInjectBuiltins(CompilingNodeTransformer):
    step = "Injecting builtin functions"

    def visit_Module(self, node: TypedModule) -> TypedModule:
        additional_assigns = []
        for b in PythonBuiltIn:
            typ = PythonBuiltInTypes[b]
            if not isinstance(b.value, plt.AST):
                # skip polymorphic functions
                continue
            additional_assigns.append(
                TypedAssign(
                    targets=[TypedName(id=b.name, typ=typ, ctx=Store())],
                    value=RawPlutoExpr(typ=typ, expr=force_params(deepcopy(b.value))),
                )
            )
        md = copy(node)
        # prepend all builtin definitions
        md.body = additional_assigns + node.body
        return md



from ..type_impls import *


def test_record_type_order():
    A = RecordType(Record("A", "A", 0, [("foo", IntegerInstanceType)]))
    B = RecordType(Record("B", "B", 1, [("bar", IntegerInstanceType)]))
    C = RecordType(Record("C", "C", 2, [("baz", IntegerInstanceType)]))
    a = A
    b = B
    c = C

    assert a >= a
    assert not a >= b
    assert not b >= a
    assert not a >= c
    assert not c >= a
    assert not b >= c
    assert not c >= b

    A = RecordType(Record("A", "A", 0, [("foo", IntegerInstanceType)]))
    B = RecordType(
        Record(
            "B", "B", 0, [("foo", IntegerInstanceType), ("bar", IntegerInstanceType)]
        )
    )
    C = RecordType(Record("C", "C", 0, [("foo", InstanceType(AnyType()))]))
    assert not A >= B
    assert not C >= B
    assert C >= A


def test_union_type_order():
    A = RecordType(Record("A", "A", 0, [("foo", IntegerInstanceType)]))
    B = RecordType(Record("B", "B", 1, [("bar", IntegerInstanceType)]))
    C = RecordType(Record("C", "C", 2, [("baz", IntegerInstanceType)]))
    abc = UnionType([A, B, C])
    ab = UnionType([A, B])
    a = A
    c = C

    assert a >= a
    assert ab >= a
    assert not a >= ab
    assert abc >= ab
    assert not ab >= abc
    assert not c >= a
    assert not a >= c
    assert abc >= c
    assert not ab >= c



import re
from copy import copy
from typing import Optional
from enum import Enum

from ..type_inference import INITIAL_SCOPE
from ..util import CompilingNodeTransformer
from ..typed_ast import *

"""
Injects the integrity checker function if it is imported
"""

FunctionName = "check_integrity"


class IntegrityCheckImpl(PolymorphicFunction):
    def type_from_args(self, args: typing.List[Type]) -> FunctionType:
        assert (
            len(args) == 1
        ), f"'integrity_check' takes only one argument, but {len(args)} were given"
        typ = args[0]
        assert isinstance(typ, InstanceType), "Can only check integrity of instances"
        assert any(
            isinstance(typ.typ, t) for t in (RecordType, UnionType)
        ), "Can only check integrity of PlutusData and Union types"
        return FunctionType(args, NoneInstanceType)

    def impl_from_args(self, args: typing.List[Type]) -> plt.AST:
        arg = args[0]
        assert isinstance(arg, InstanceType), "Can only check integrity of instances"
        return OLambda(
            ["x"],
            plt.Ite(
                plt.EqualsData(
                    OVar("x"),
                    plt.Apply(arg.typ.copy_only_attributes(), OVar("x")),
                ),
                plt.Unit(),
                plt.TraceError("ValueError: datum integrity check failed"),
            ),
        )


class RewriteImportIntegrityCheck(CompilingNodeTransformer):
    step = "Resolving imports and usage of integrity check"

    def visit_ImportFrom(self, node: ImportFrom) -> Optional[AST]:
        if node.module != "opshin.std.integrity":
            return node
        for n in node.names:
            assert (
                n.name == FunctionName
            ), "Imports something other from the integrity check than the integrity check builtin"
            renamed = n.asname if n.asname is not None else n.name
            INITIAL_SCOPE[renamed] = InstanceType(
                PolymorphicFunctionType(IntegrityCheckImpl())
            )
        return None



from typing import Tuple

import hypothesis
import hypothesis.strategies as hst

from opshin.std import bitmap as oc_bitmap


@hypothesis.given(hst.integers(min_value=0, max_value=1000))
def test_init_bitmap(x: int):
    assert (
        len(oc_bitmap.init_bitmap(x)) * 8 >= x
    ), "Insufficient size of initialized bitmap"


@hst.composite
def bytes_and_index(draw):
    xs = draw(hst.binary(min_size=1))
    i = draw(hst.integers(min_value=0, max_value=len(xs)))
    return (xs, i)


@hypothesis.given(bytes_and_index())
def test_test_bitmap(p: Tuple[bytes, int]):
    a, i = p
    ith_bit = int.from_bytes(a, "big") & (1 << ((len(a) * 8) - i - 1))
    assert oc_bitmap.test_bitmap(a, i) == (ith_bit != 0), "Invalid isset check"


@hypothesis.given(bytes_and_index(), hst.booleans())
def test__set_bitmap(p: Tuple[bytes, int], v: bool):
    a, i = p
    if v:
        set_bit = int.from_bytes(a, "big") | (1 << ((len(a) * 8) - i - 1))
    else:
        set_bit = int.from_bytes(a, "big") & (~(1 << ((len(a) * 8) - i - 1)))
    set_bit = set_bit.to_bytes(len(a), byteorder="big")
    assert oc_bitmap._set_bitmap(a, i, v) == set_bit, "Set bit incorrectly"


@hypothesis.given(bytes_and_index())
def test_set_bitmap(p: Tuple[bytes, int]):
    a, i = p
    set_bit = int.from_bytes(a, "big") | (1 << ((len(a) * 8) - i - 1))
    set_bit = set_bit.to_bytes(len(a), byteorder="big")
    assert oc_bitmap.set_bitmap(a, i) == set_bit, "Set bit incorrectly"


@hypothesis.given(bytes_and_index())
def test_reset_bitmap(p: Tuple[bytes, int]):
    a, i = p
    set_bit = int.from_bytes(a, "big") & (~(1 << ((len(a) * 8) - i - 1)))
    set_bit = set_bit.to_bytes(len(a), byteorder="big")
    assert oc_bitmap.reset_bitmap(a, i) == set_bit, "Reset bit incorrectly"


@hypothesis.given(bytes_and_index())
def test_flip_bitmap(p: Tuple[bytes, int]):
    a, i = p
    set_bit = int.from_bytes(a, "big") ^ (1 << ((len(a) * 8) - i - 1))
    set_bit = set_bit.to_bytes(len(a), byteorder="big")
    assert oc_bitmap.flip_bitmap(a, i) == set_bit, "Flipped bit incorrectly"


@hypothesis.given(hst.binary())
def test_size_bitmap(x: bytes):
    assert len(x) * 8 == oc_bitmap.size_bitmap(x), "Incorrect size of bitmap"


@hypothesis.given(hst.binary())
def test_all_bitmap(x: bytes):
    assert oc_bitmap.all_bitmap(x) == all(
        b == 0xFF for b in x
    ), "All is incorrect for bitmap"


@hypothesis.given(hst.binary())
def test_any_bitmap(x: bytes):
    assert oc_bitmap.any_bitmap(x) == any(
        b != 0x00 for b in x
    ), "Any is incorrect for bitmap"


@hypothesis.given(hst.binary())
def test_none_bitmap(x: bytes):
    assert oc_bitmap.none_bitmap(x) == all(
        b == 0x00 for b in x
    ), "None is incorrect for bitmap"


@hypothesis.given(bytes_and_index())
def test_flip_roundtrip(p: Tuple[bytes, int]):
    a, i = p
    assert (
        oc_bitmap.flip_bitmap(oc_bitmap.flip_bitmap(a, i), i) == a
    ), "Flipped bit incorrectly"


@hypothesis.given(bytes_and_index())
def test_set_test(p: Tuple[bytes, int]):
    a, i = p
    assert oc_bitmap.test_bitmap(oc_bitmap.set_bitmap(a, i), i), "Set bit incorrectly"


@hypothesis.given(bytes_and_index())
def test_reset_test(p: Tuple[bytes, int]):
    a, i = p
    assert not oc_bitmap.test_bitmap(
        oc_bitmap.reset_bitmap(a, i), i
    ), "Reset bit incorrectly"



import inspect

import argparse
import logging
import tempfile

import cbor2
import enum
import importlib
import json
import pathlib
import sys
import typing
import ast

import pycardano
from pycardano import PlutusData, RawPlutusData

import pluthon
import uplc
import uplc.ast

from . import (
    compiler,
    builder,
    prelude,
    __version__,
    __copyright__,
    Purpose,
    PlutusContract,
)
from .util import CompilerError, data_from_json, OPSHIN_LOG_HANDLER
from .prelude import ScriptContext
from .compiler_config import *


class Command(enum.Enum):
    compile_pluto = "compile_pluto"
    compile = "compile"
    eval = "eval"
    parse = "parse"
    eval_uplc = "eval_uplc"
    build = "build"
    lint = "lint"


def parse_uplc_param(param: str):
    if param.startswith("{"):
        try:
            return uplc.ast.data_from_json_dict(json.loads(param))
        except json.JSONDecodeError as e:
            raise ValueError(
                f"Invalid parameter for contract passed, expected json value, got {param}"
            ) from e
    else:
        try:
            return uplc.ast.data_from_cbor(bytes.fromhex(param))
        except ValueError as e:
            raise ValueError(
                "Expected hexadecimal CBOR representation of plutus datum but could not transform hex string to bytes."
            ) from e


def parse_plutus_param(annotation, param: str):
    if param.startswith("{"):
        try:
            param_dict = json.loads(param)
        except json.JSONDecodeError as e:
            raise ValueError(
                f"Invalid parameter for contract passed, expected json value, got {param}"
            ) from e
        return plutus_data_from_json(annotation, param_dict)
    else:
        try:
            param_bytes = bytes.fromhex(param)
        except ValueError as e:
            raise ValueError(
                "Expected hexadecimal CBOR representation of plutus datum but could not transform hex string to bytes."
            ) from e
        return plutus_data_from_cbor(annotation, param_bytes)


def plutus_data_from_json(annotation: typing.Type, x: dict):
    try:
        if annotation == int:
            return int(x["int"])
        if annotation == bytes:
            return bytes.fromhex(x["bytes"])
        if annotation is None:
            return None
        if isinstance(annotation, typing._GenericAlias):
            # Annotation is a List or Dict
            if annotation._name == "List":
                annotation_ann = annotation.__dict__["__args__"][0]
                return [plutus_data_from_json(annotation_ann, k) for k in x["list"]]
            if annotation._name == "Dict":
                annotation_key, annotation_val = annotation.__dict__["__args__"]
                return {
                    plutus_data_from_json(
                        annotation_key, d["k"]
                    ): plutus_data_from_json(annotation_val, d["v"])
                    for d in x["map"]
                }
            if annotation.__origin__ == typing.Union:
                for ann in annotation.__dict__["__args__"]:
                    try:
                        return plutus_data_from_json(ann, x)
                    except pycardano.DeserializeException:
                        pass
                raise ValueError(
                    f"Could not find matching type for {x} in {annotation}"
                )
            if annotation == pycardano.Datum:
                if "int" in x:
                    return int(x["int"])
                if "bytes" in x:
                    return bytes.fromhex(x["bytes"])
                if "constructor" in x:
                    return pycardano.RawCBOR(
                        uplc.ast.plutus_cbor_dumps(uplc.ast.data_from_json_dict(x))
                    )
                if "list" in x:
                    return [
                        plutus_data_from_json(pycardano.Datum, k) for k in x["list"]
                    ]
                if "map" in x:
                    return {
                        plutus_data_from_json(
                            pycardano.Datum, d["k"]
                        ): plutus_data_from_json(pycardano.Datum, d["v"])
                        for d in x["map"]
                    }
        if issubclass(annotation, pycardano.PlutusData):
            return annotation.from_dict(x)
    except (KeyError, ValueError):
        raise ValueError(
            f"Annotation {annotation} does not match provided plutus datum {json.dumps(x)}"
        )


def plutus_data_from_cbor(annotation: typing.Type, x: bytes):
    try:
        if annotation in (int, bytes):
            return cbor2.loads(x)
        if annotation is None:
            return None
        if isinstance(annotation, typing._GenericAlias):
            # Annotation is a List or Dict
            if annotation.__origin__ == list:
                annotation_ann = annotation.__dict__["__args__"][0]
                return [
                    plutus_data_from_cbor(annotation_ann, cbor2.dumps(k))
                    for k in cbor2.loads(x)
                ]
            if annotation.__origin__ == dict:
                annotation_key, annotation_val = annotation.__dict__["__args__"]
                return {
                    plutus_data_from_cbor(
                        annotation_key, cbor2.dumps(k)
                    ): plutus_data_from_cbor(annotation_val, v)
                    for k, v in cbor2.loads(x).items()
                }
            if annotation.__origin__ == typing.Union:
                for ann in annotation.__dict__["__args__"]:
                    try:
                        return plutus_data_from_cbor(ann, x)
                    except pycardano.DeserializeException:
                        pass
                raise ValueError(
                    f"Could not find matching type for {x.hex()} in {annotation}"
                )
        if issubclass(annotation, pycardano.PlutusData):
            return annotation.from_cbor(x)
    except (KeyError, ValueError):
        raise ValueError(
            f"Annotation {annotation} does not match provided plutus datum {x.hex()}"
        )


def check_params(
    command: Command,
    purpose: Purpose,
    validator_args,
    return_type,
    validator_params,
    force_three_params=False,
):
    num_onchain_params = (
        3
        if purpose == Purpose.spending or force_three_params or purpose == Purpose.any
        else 2
    )
    onchain_params = validator_args[-num_onchain_params:]
    param_types = validator_args[:-num_onchain_params]
    if purpose == Purpose.any:
        # The any purpose does not do any checks. Use only if you know what you are doing
        return onchain_params, param_types
    # expect the validator to return None
    assert (
        return_type is None or return_type == prelude.Anything
    ), f"Expected contract to return None, but returns {return_type}"

    required_onchain_parameters = 3 if purpose == Purpose.spending else 2
    if force_three_params:
        datum_type = onchain_params[0][1]
        assert (
            (
                typing.get_origin(datum_type) == typing.Union
                and prelude.Nothing in typing.get_args(datum_type)
            )
            or datum_type == prelude.Anything
            or datum_type == prelude.Nothing
        ), f"Expected contract to accept Nothing or Anything as datum since it forces three parameters, but got {datum_type}"

    assert (
        len(onchain_params) == required_onchain_parameters
    ), f"""\
{purpose.value.capitalize()} validator must expect {required_onchain_parameters} parameters at evaluation (on-chain), but was specified to have {len(onchain_params)}.
Make sure the validator expects parameters {'datum, ' if purpose == Purpose.spending else ''}redeemer and script context."""

    if command in (Command.eval, Command.eval_uplc):
        assert len(validator_params) == len(param_types) + len(
            onchain_params
        ), f"{purpose.value.capitalize()} validator expects {len(param_types) + len(onchain_params)} parameters for evaluation, but only got {len(validator_params)}."
    assert (
        onchain_params[-1][1] == ScriptContext
    ), f"Last parameter of the validator is always ScriptContext, but is {onchain_params[-1][1].__name__} here."
    return onchain_params, param_types


def perform_command(args):
    # generate the compiler config
    compiler_config = DEFAULT_CONFIG
    compiler_config = compiler_config.update(OPT_CONFIGS[args.opt_level])
    overrides = {}
    for k in ARGPARSE_ARGS.keys():
        if getattr(args, k) is not None:
            overrides[k] = getattr(args, k)
    compiler_config = compiler_config.update(CompilationConfig(**overrides))
    # configure logging
    if args.verbose:
        OPSHIN_LOG_HANDLER.setLevel(logging.DEBUG)

    # execute the command
    command = Command(args.command)
    purpose = Purpose(args.purpose)
    if purpose == Purpose.lib:
        assert (
            not compiler_config.remove_dead_code
        ), "Libraries must have dead code removal disabled (-fno-remove-dead-code)"
    input_file = args.input_file if args.input_file != "-" else sys.stdin
    # read and import the contract
    with open(input_file, "r") as f:
        source_code = f.read()
    with tempfile.TemporaryDirectory(prefix="build") as tmpdir:
        tmp_input_file = pathlib.Path(tmpdir).joinpath("__tmp_opshin.py")
        with tmp_input_file.open("w") as fp:
            fp.write(source_code)
        sys.path.append(str(pathlib.Path(tmp_input_file).parent.absolute()))
        sc = importlib.import_module(pathlib.Path(tmp_input_file).stem)
        sys.path.pop()
    # load the passed parameters if not a lib
    if purpose == Purpose.lib:
        assert not args.args, "Can not pass arguments to a library"
        parsed_params = []
        uplc_params = []
    else:
        try:
            argspec = inspect.signature(sc.validator)
        except AttributeError:
            raise AssertionError(
                f"Contract has no function called 'validator'. Make sure the compiled contract contains one function called 'validator' or {command.value} using `opshin {command.value} lib {str(input_file)}`."
            )
        annotations = [
            (x.name, x.annotation or prelude.Anything)
            for x in argspec.parameters.values()
        ]
        return_annotation = (
            argspec.return_annotation
            if argspec.return_annotation not in (None, argspec.empty)
            else prelude.Anything
        )
        parsed_params = []
        uplc_params = []
        for i, (c, a) in enumerate(zip(annotations, args.args)):
            uplc_param = parse_uplc_param(a)
            uplc_params.append(uplc_param)
            param = parse_plutus_param(c[1], a)
            parsed_params.append(param)
        onchain_params, param_types = check_params(
            command,
            purpose,
            annotations,
            return_annotation,
            parsed_params,
            compiler_config.force_three_params,
        )

    if command == Command.eval:
        assert purpose != Purpose.lib, "Can not evaluate a library"
        print("Starting execution")
        print("------------------")
        try:
            ret = sc.validator(*parsed_params)
        except Exception as e:
            print(f"Exception of type {type(e).__name__} caused")
            ret = e
        print("------------------")
        print(ret)

    source_ast = compiler.parse(source_code, filename=input_file)

    if command == Command.parse:
        print("Parsed successfully.")
        return

    try:
        code = compiler.compile(
            source_ast,
            filename=input_file,
            validator_function_name="validator" if purpose != Purpose.lib else None,
            # do not remove dead code when compiling a library - none of the code will be used
            config=compiler_config,
        )
    except CompilerError as c:
        # Generate nice error message from compiler error
        if not isinstance(c.node, ast.Module):
            source_seg = ast.get_source_segment(source_code, c.node)
            start_line = c.node.lineno - 1
            end_line = start_line + len(source_seg.splitlines())
            source_lines = "\n".join(source_code.splitlines()[start_line:end_line])
            pos_in_line = source_lines.find(source_seg)
        else:
            start_line = 0
            pos_in_line = 0
            source_lines = source_code.splitlines()[0]

        overwrite_syntaxerror = (
            len("SyntaxError: ") * "\b" if command != Command.lint else ""
        )
        err = SyntaxError(
            f"""\
{overwrite_syntaxerror}{c.orig_err.__class__.__name__}: {c.orig_err}
Note that opshin errors may be overly restrictive as they aim to prevent code with unintended consequences.
""",
            (
                args.input_file,
                start_line + 1,
                pos_in_line,
                source_lines,
            ),
            # we remove chaining so that users to not see the internal trace back,
        )
        err.orig_err = c.orig_err
        raise err from None

    if command == Command.compile_pluto:
        print(code.dumps())
        return
    code = pluthon.compile(code, config=compiler_config)

    # apply parameters from the command line to the contract (instantiates parameterized contract!)
    code = code.term
    # UPLC lambdas may only take one argument at a time, so we evaluate by repeatedly applying
    for d in uplc_params:
        code = uplc.ast.Apply(code, d)
    code = uplc.ast.Program((1, 0, 0), code)

    if command == Command.compile:
        print(code.dumps())
        return

    if command == Command.build:
        if args.output_directory == "":
            if args.input_file == "-":
                print(
                    "Please supply an output directory if no input file is specified."
                )
                exit(-1)
            target_dir = pathlib.Path("build") / pathlib.Path(input_file).stem
        else:
            target_dir = pathlib.Path(args.output_directory)
        built_code = builder._build(code)
        if purpose == Purpose.lib:
            script_arts = PlutusContract(
                built_code,
            )
        else:
            script_arts = PlutusContract(
                built_code,
                datum_type=onchain_params[0] if len(onchain_params) == 3 else None,
                redeemer_type=onchain_params[1 if len(onchain_params) == 3 else 0],
                parameter_types=param_types,
                purpose=(purpose,),
            )
        script_arts.dump(target_dir)

        print(f"Wrote script artifacts to {target_dir}/")
        return
    if command == Command.eval_uplc:
        print("Starting execution")
        print("------------------")
        assert isinstance(code, uplc.ast.Program)
        try:
            ret = uplc.dumps(uplc.eval(code))
        except Exception as e:
            print("An exception was raised")
            ret = e
        print("------------------")
        print(ret)


def parse_args():
    a = argparse.ArgumentParser(
        description="An evaluator and compiler from python into UPLC. Translate imperative programs into functional quasi-assembly. Flags allow setting fine-grained compiler options. All flags can be turned off via -fno-<flag>."
    )
    a.add_argument(
        "command",
        type=str,
        choices=Command.__members__.keys(),
        help="The command to execute on the input file.",
        default="eval",
        nargs="?",
    )
    a.add_argument(
        "purpose",
        type=str,
        choices=Purpose.__members__.keys(),
        help="The intended script purpose. Determines the number of on-chain parameters "
        "(spending = 3, minting, rewarding, certifying = 2, any = no checks). "
        "This allows the compiler to check whether the correct amount of parameters was passed during compilation.",
        default="any",
        nargs="?",
    )
    a.add_argument(
        "input_file", type=str, help="The input program to parse. Set to - for stdin."
    )
    a.add_argument(
        "-o",
        "--output-directory",
        default="",
        type=str,
        help="The output directory for artefacts of the build command. Defaults to the filename of the compiled contract. of the compiled contract.",
    )
    a.add_argument(
        "args",
        nargs="*",
        default=[],
        help="Input parameters for the validator (parameterizes the contract for compile/build). Either json or CBOR notation.",
    )
    a.add_argument(
        "--output-format-json",
        action="store_true",
        help="Changes the output of the Linter to a json format.",
    )
    a.add_argument(
        "--version",
        action="version",
        version=f"opshin {__version__} {__copyright__}",
    )
    a.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Enable verbose logging.",
    )
    a.add_argument(
        "--recursion-limit",
        default=sys.getrecursionlimit(),
        help="Modify the recursion limit (necessary for larger UPLC programs)",
        type=int,
    )
    for k, v in ARGPARSE_ARGS.items():
        alts = v.pop("__alts__", [])
        a.add_argument(
            f"-f{k.replace('_', '-')}",
            *alts,
            **v,
            action="store_true",
            dest=k,
            default=None,
        )
        a.add_argument(
            f"-fno-{k.replace('_', '-')}",
            action="store_false",
            help=argparse.SUPPRESS,
            dest=k,
            default=None,
        )
    a.add_argument(
        f"-O",
        type=int,
        help=f"Optimization level from 0 (no optimization) to 3 (aggressive optimization, removes traces). Defaults to 1.",
        default=1,
        choices=range(len(OPT_CONFIGS)),
        dest="opt_level",
    )
    return a.parse_args()


def main():
    args = parse_args()
    sys.setrecursionlimit(args.recursion_limit)
    if Command(args.command) != Command.lint:
        OPSHIN_LOG_HANDLER.setFormatter(
            logging.Formatter(
                f"%(levelname)s for {args.input_file}:%(lineno)d %(message)s"
            )
        )
        perform_command(args)
    else:
        OPSHIN_LOG_HANDLER.stream = sys.stdout
        if args.output_format_json:
            OPSHIN_LOG_HANDLER.setFormatter(
                logging.Formatter(
                    '{"line":%(lineno)d,"column":%(col_offset)d,"error_class":"%(levelname)s","message":"%(message)s"}'
                )
            )
        else:
            OPSHIN_LOG_HANDLER.setFormatter(
                logging.Formatter(
                    args.input_file
                    + ":%(lineno)d:%(col_offset)d:%(levelname)s: %(message)s"
                )
            )

        try:
            perform_command(args)
        except Exception as e:
            error_class_name = e.__class__.__name__
            message = str(e)
            if isinstance(e, SyntaxError):
                start_line = e.lineno
                pos_in_line = e.offset
                if hasattr(e, "orig_err"):
                    error_class_name = e.orig_err.__class__.__name__
                    message = str(e.orig_err)
            else:
                start_line = 1
                pos_in_line = 1
            if args.output_format_json:
                print(
                    convert_linter_to_json(
                        line=start_line,
                        column=pos_in_line,
                        error_class=error_class_name,
                        message=message,
                    )
                )
            else:
                print(
                    f"{args.input_file}:{start_line}:{pos_in_line}: {error_class_name}: {message}"
                )


def convert_linter_to_json(
    line: int,
    column: int,
    error_class: str,
    message: str,
):
    # output in lists
    return json.dumps(
        [
            {
                "line": line,
                "column": column,
                "error_class": error_class,
                "message": message,
            }
        ]
    )


if __name__ == "__main__":
    main()



"""
An implementation of fractions in opshin
This does not maintain smallest possible notation invariants for the sake of efficiency
- the user has full control over when to normalize the fractions and should do so using norm_fraction
"""

from dataclasses import dataclass
from pycardano import Datum as Anything, PlutusData
from typing import Dict, List, Union

from opshin.std.math import *


@dataclass(unsafe_hash=True)
class Fraction(PlutusData):
    CONSTR_ID = 1
    numerator: int
    denominator: int


def add_fraction(a: Fraction, b: Fraction) -> Fraction:
    """returns a + b"""
    return Fraction(
        (a.numerator * b.denominator) + (b.numerator * a.denominator),
        a.denominator * b.denominator,
    )


def neg_fraction(a: Fraction) -> Fraction:
    """returns -a"""
    return Fraction(-a.numerator, a.denominator)


def sub_fraction(a: Fraction, b: Fraction) -> Fraction:
    """returns a - b"""
    return add_fraction(a, neg_fraction(b))


def mul_fraction(a: Fraction, b: Fraction) -> Fraction:
    """returns a * b"""
    return Fraction(a.numerator * b.numerator, a.denominator * b.denominator)


def div_fraction(a: Fraction, b: Fraction) -> Fraction:
    """returns a / b"""
    return Fraction(a.numerator * b.denominator, a.denominator * b.numerator)


def _norm_signs_fraction(a: Fraction) -> Fraction:
    """Restores the invariant that the denominator is > 0"""
    return Fraction(sign(a.denominator) * a.numerator, abs(a.denominator))


def _norm_gcd_fraction(a: Fraction) -> Fraction:
    """Restores the invariant that num/denom are in the smallest possible denomination"""
    g = gcd(a.numerator, a.denominator)
    return Fraction(a.numerator // g, a.denominator // g)


def norm_fraction(a: Fraction) -> Fraction:
    """Restores the invariant that num/denom are in the smallest possible denomination and denominator > 0"""
    return _norm_gcd_fraction(_norm_signs_fraction(a))


def ge_fraction(a: Fraction, b: Fraction) -> bool:
    """returns a >= b"""
    if a.denominator * b.denominator >= 0:
        res = a.numerator * b.denominator >= a.denominator * b.numerator
    else:
        res = a.numerator * b.denominator <= a.denominator * b.numerator
    return res


def le_fraction(a: Fraction, b: Fraction) -> bool:
    """returns a <= b"""
    if a.denominator * b.denominator >= 0:
        res = a.numerator * b.denominator <= a.denominator * b.numerator
    else:
        res = a.numerator * b.denominator >= a.denominator * b.numerator
    return res


def eq_fraction(a: Fraction, b: Fraction) -> bool:
    """returns a == b"""
    return a.numerator * b.denominator == a.denominator * b.numerator


def lt_fraction(a: Fraction, b: Fraction) -> bool:
    """returns a < b"""
    if a.denominator * b.denominator >= 0:
        res = a.numerator * b.denominator < a.denominator * b.numerator
    else:
        res = a.numerator * b.denominator > a.denominator * b.numerator
    return res


def gt_fraction(a: Fraction, b: Fraction) -> bool:
    """returns a > b"""
    if a.denominator * b.denominator >= 0:
        res = a.numerator * b.denominator > a.denominator * b.numerator
    else:
        res = a.numerator * b.denominator < a.denominator * b.numerator
    return res


def floor_fraction(a: Fraction) -> int:
    return a.numerator // a.denominator


def ceil_fraction(a: Fraction) -> int:
    return (a.numerator + a.denominator - sign(a.denominator)) // a.denominator



from _ast import Name, Store, ClassDef, FunctionDef, Load
from collections import defaultdict
from copy import copy, deepcopy
import logging

import typing

import ast
from dataclasses import dataclass

import pycardano
from frozendict import frozendict
from frozenlist2 import frozenlist

import uplc.ast as uplc
import pluthon as plt
from hashlib import sha256

OPSHIN_LOGGER = logging.getLogger("opshin")
OPSHIN_LOG_HANDLER = logging.StreamHandler()
OPSHIN_LOGGER.addHandler(OPSHIN_LOG_HANDLER)


class FileContextFilter(logging.Filter):
    """
    This is a filter which injects contextual information into the log.

    The information is about the currently inspected AST node.
    The information needs to be updated inside the NodeTransformer and NodeVisitor classes.
    """

    file_name = "unknown"
    node: ast.AST = None

    def filter(self, record):

        if self.node is None:
            record.lineno = 1
            record.col_offset = 0
            record.end_lineno = 1
            record.end_col_offset = 0
        else:
            record.lineno = self.node.lineno
            record.col_offset = self.node.col_offset
            record.end_lineno = self.node.end_lineno
            record.end_col_offset = self.node.end_col_offset
        return True


OPSHIN_LOG_CONTEXT_FILTER = FileContextFilter()
OPSHIN_LOG_HANDLER.addFilter(OPSHIN_LOG_CONTEXT_FILTER)


def distinct(xs: list):
    """Returns true iff the list consists of distinct elements"""
    return len(xs) == len(set(xs))


class TypedNodeTransformer(ast.NodeTransformer):
    def visit(self, node):
        """Visit a node."""
        OPSHIN_LOG_CONTEXT_FILTER.node = node
        node_class_name = node.__class__.__name__
        if node_class_name.startswith("Typed"):
            node_class_name = node_class_name[len("Typed") :]
        method = "visit_" + node_class_name
        visitor = getattr(self, method, self.generic_visit)
        return visitor(node)


class TypedNodeVisitor(ast.NodeVisitor):
    def visit(self, node):
        """Visit a node."""
        OPSHIN_LOG_CONTEXT_FILTER.node = node
        node_class_name = node.__class__.__name__
        if node_class_name.startswith("Typed"):
            node_class_name = node_class_name[len("Typed") :]
        method = "visit_" + node_class_name
        visitor = getattr(self, method, self.generic_visit)
        return visitor(node)


class CompilerError(Exception):
    def __init__(self, orig_err: Exception, node: ast.AST, compilation_step: str):
        self.orig_err = orig_err
        self.node = node
        self.compilation_step = compilation_step


class CompilingNodeTransformer(TypedNodeTransformer):
    step = "Node transformation"

    def visit(self, node):
        OPSHIN_LOG_CONTEXT_FILTER.node = node
        try:
            return super().visit(node)
        except Exception as e:
            if isinstance(e, CompilerError):
                raise e
            raise CompilerError(e, node, self.step)


class NoOp(CompilingNodeTransformer):
    """A variation of the Compiling Node transformer that performs no changes"""

    pass


class CompilingNodeVisitor(TypedNodeVisitor):
    step = "Node visiting"

    def visit(self, node):
        try:
            return super().visit(node)
        except Exception as e:
            if isinstance(e, CompilerError):
                raise e
            raise CompilerError(e, node, self.step)


def data_from_json(j: typing.Dict[str, typing.Any]) -> uplc.PlutusData:
    if "bytes" in j:
        return uplc.PlutusByteString(bytes.fromhex(j["bytes"]))
    if "int" in j:
        return uplc.PlutusInteger(int(j["int"]))
    if "list" in j:
        return uplc.PlutusList(frozenlist(list(map(data_from_json, j["list"]))))
    if "map" in j:
        return uplc.PlutusMap(
            frozendict(
                {data_from_json(d["k"]): data_from_json(d["v"]) for d in j["map"]}
            )
        )
    if "constructor" in j and "fields" in j:
        return uplc.PlutusConstr(
            j["constructor"], frozenlist(list(map(data_from_json, j["fields"])))
        )
    raise NotImplementedError(f"Unknown datum representation {j}")


def datum_to_cbor(d: pycardano.Datum) -> bytes:
    return pycardano.PlutusData.to_cbor(d)


def datum_to_json(d: pycardano.Datum) -> str:
    return pycardano.PlutusData.to_json(d)


def custom_fix_missing_locations(node, parent=None):
    """
    Works like ast.fix_missing_location but forces it onto everything
    """

    def _fix(node, lineno, col_offset, end_lineno, end_col_offset):
        if getattr(node, "lineno", None) is None:
            node.lineno = lineno
        else:
            lineno = node.lineno
        if getattr(node, "end_lineno", None) is None:
            node.end_lineno = end_lineno
        else:
            end_lineno = node.end_lineno
        if getattr(node, "col_offset", None) is None:
            node.col_offset = col_offset
        else:
            col_offset = node.col_offset
        if getattr(node, "end_col_offset", None) is None:
            node.end_col_offset = end_col_offset
        else:
            end_col_offset = node.end_col_offset
        for child in ast.iter_child_nodes(node):
            _fix(child, lineno, col_offset, end_lineno, end_col_offset)

    lineno, col_offset, end_lineno, end_col_offset = (
        getattr(parent, "lineno", 1),
        getattr(parent, "col_offset", 0),
        getattr(parent, "end_lineno", 1),
        getattr(parent, "end_col_offset", 0),
    )
    _fix(node, lineno, col_offset, end_lineno, end_col_offset)
    return node


_patterns_cached = {}


def make_pattern(structure: plt.AST) -> plt.Pattern:
    """Creates a shared pattern from the given lambda, cached so that it is re-used in subsequent calls"""
    structure_serialized = structure.dumps()
    if _patterns_cached.get(structure_serialized) is None:
        # @dataclass
        # class AdHocPattern(plt.Pattern):

        #     def compose(self):
        #         return structure
        AdHocPattern = type(
            f"AdHocPattern_{sha256(structure_serialized.encode()).digest().hex()}",
            (plt.Pattern,),
            {"compose": lambda self: deepcopy(structure)},
        )
        AdHocPattern = dataclass(AdHocPattern)

        _patterns_cached[structure_serialized] = AdHocPattern()
    return deepcopy(_patterns_cached[structure_serialized])


def patternize(method):
    def wrapped(*args, **kwargs):
        return make_pattern(method(*args, **kwargs))

    return wrapped


def force_params(lmd: plt.Lambda) -> plt.Lambda:
    if isinstance(lmd, plt.Lambda):
        return plt.Lambda(
            lmd.vars, plt.Let([(v, plt.Force(plt.Var(v))) for v in lmd.vars], lmd.term)
        )
    if isinstance(lmd, plt.Pattern):
        return make_pattern(force_params(lmd.compose()))


class NameWriteCollector(CompilingNodeVisitor):
    step = "Collecting variables that are written"

    def __init__(self):
        self.written = defaultdict(int)

    def visit_Name(self, node: Name) -> None:
        if isinstance(node.ctx, Store):
            self.written[node.id] += 1

    def visit_ClassDef(self, node: ClassDef):
        # ignore the content (i.e. attribute names) of class definitions
        self.written[node.name] += 1
        pass

    def visit_FunctionDef(self, node: FunctionDef):
        # ignore the type hints of function arguments
        self.written[node.name] += 1
        for a in node.args.args:
            self.written[a.arg] += 1
        for s in node.body:
            self.visit(s)


def written_vars(node):
    """
    Returns all variable names written to in this node
    """
    collector = NameWriteCollector()
    collector.visit(node)
    return sorted(collector.written.keys())


class NameReadCollector(CompilingNodeVisitor):
    step = "Collecting variables that are read"

    def __init__(self):
        self.read = defaultdict(int)

    def visit_AnnAssign(self, node) -> None:
        # ignore annotations of variables
        self.visit(node.value)
        self.visit(node.target)

    def visit_FunctionDef(self, node) -> None:
        # ignore annotations of paramters and return
        for b in node.body:
            self.visit(b)

    def visit_Name(self, node: Name) -> None:
        if isinstance(node.ctx, Load):
            self.read[node.id] += 1

    def visit_ClassDef(self, node: ClassDef):
        # ignore the content (i.e. attribute names) of class definitions
        pass


def read_vars(node):
    """
    Returns all variable names read to in this node
    """
    collector = NameReadCollector()
    collector.visit(node)
    return sorted(collector.read.keys())


def all_vars(node):
    return sorted(set(read_vars(node) + written_vars(node)))


def externally_bound_vars(node: FunctionDef):
    """A superset of the variables bound from an outer scope"""
    return sorted(set(read_vars(node)) - set(written_vars(node)) - {"isinstance"})


def opshin_name_scheme_compatible_varname(n: str):
    return f"1{n}"


def OVar(name: str):
    return plt.Var(opshin_name_scheme_compatible_varname(name))


def OLambda(names: typing.List[str], term: plt.AST):
    return plt.Lambda([opshin_name_scheme_compatible_varname(x) for x in names], term)


def OLet(bindings: typing.List[typing.Tuple[str, plt.AST]], term: plt.AST):
    return plt.Let(
        [(opshin_name_scheme_compatible_varname(n), t) for n, t in bindings], term
    )


def SafeLambda(vars: typing.List[str], term: plt.AST) -> plt.Lambda:
    if not vars:
        return plt.Lambda(["0_"], term)
    return plt.Lambda(vars, term)


def SafeOLambda(vars: typing.List[str], term: plt.AST) -> plt.Lambda:
    if not vars:
        return OLambda(["0_"], term)
    return OLambda(vars, term)


def SafeApply(term: plt.AST, *vars: typing.List[plt.AST]) -> plt.Apply:
    if not vars:
        return plt.Apply(term, plt.Delay(plt.Unit()))
    return plt.Apply(term, *vars)



from .type_impls import *


class TypedAST(AST):
    typ: Type


class typedexpr(TypedAST, expr):
    def typechecks(self) -> typing.Dict[str, Type]:
        """Successful typechecks if this expression evaluates to True"""
        return {}


class typedstmt(TypedAST, stmt):
    # Statements always have type None
    typ = NoneInstanceType


class typedarg(TypedAST, arg):
    pass


class typedarguments(TypedAST, arguments):
    args: typing.List[typedarg]
    vararg: typing.Union[typedarg, None]
    kwonlyargs: typing.List[typedarg]
    kw_defaults: typing.List[typing.Union[typedexpr, None]]
    kwarg: typing.Union[typedarg, None]
    defaults: typing.List[typedexpr]


class TypedModule(typedstmt, Module):
    body: typing.List[typedstmt]


class TypedFunctionDef(typedstmt, FunctionDef):
    body: typing.List[typedstmt]
    args: typedarguments


class TypedIf(typedstmt, If):
    test: typedexpr
    body: typing.List[typedstmt]
    orelse: typing.List[typedstmt]


class TypedReturn(typedstmt, Return):
    value: typedexpr


class TypedExpression(typedstmt, Expression):
    body: typedexpr


class TypedCall(typedexpr, Call):
    func: typedexpr
    args: typing.List[typedexpr]


class TypedExpr(typedstmt, Expr):
    value: typedexpr


class TypedAssign(typedstmt, Assign):
    targets: typing.List[typedexpr]
    value: typedexpr


class TypedClassDef(typedstmt, ClassDef):
    class_typ: Type


class TypedAnnAssign(typedstmt, AnnAssign):
    target: typedexpr
    annotation: Type
    value: typedexpr


class TypedWhile(typedstmt, While):
    test: typedexpr
    body: typing.List[typedstmt]
    orelse: typing.List[typedstmt]


class TypedFor(typedstmt, For):
    target: typedexpr
    iter: typedexpr
    body: typing.List[typedstmt]
    orelse: typing.List[typedstmt]


class TypedPass(typedstmt, Pass):
    pass


class TypedName(typedexpr, Name):
    pass


class Typedkeyword(TypedAST, keyword):
    arg: typedexpr
    value: typedexpr


class TypedConstant(TypedAST, Constant):
    pass


class TypedTuple(typedexpr, Tuple):
    pass


class TypedList(typedexpr, List):
    pass


class typedcomprehension(typedexpr, comprehension):
    target: typedexpr
    iter: typedexpr
    ifs: typing.List[typedexpr]


class TypedListComp(typedexpr, ListComp):
    elt: typedexpr
    generators: typing.List[typedcomprehension]


class TypedDictComp(typedexpr, DictComp):
    key: typedexpr
    value: typedexpr
    generators: typing.List[typedcomprehension]


class TypedFormattedValue(typedexpr, FormattedValue):
    value: typedexpr
    conversion: int
    format_spec: typing.Optional[JoinedStr]


class TypedJoinedStr(typedexpr, JoinedStr):
    values: typing.List[typedexpr]


class TypedDict(typedexpr, Dict):
    pass


class TypedIfExp(typedstmt, IfExp):
    test: typedexpr
    body: typedexpr
    orelse: typedexpr


class TypedCompare(typedexpr, Compare):
    left: typedexpr
    ops: typing.List[cmpop]
    comparators: typing.List[typedexpr]


class TypedBinOp(typedexpr, BinOp):
    left: typedexpr
    right: typedexpr


class TypedBoolOp(typedexpr, BoolOp):
    values: typing.List[typedexpr]


class TypedUnaryOp(typedexpr, UnaryOp):
    operand: typedexpr


class TypedSubscript(typedexpr, Subscript):
    value: typedexpr


class TypedAttribute(typedexpr, Attribute):
    value: typedexpr
    pos: int


class TypedAssert(typedstmt, Assert):
    test: typedexpr
    msg: typedexpr


class RawPlutoExpr(typedexpr):
    typ: Type
    expr: plt.AST

    _attributes = ["lineno", "col_offset", "end_lineno", "end_col_offset"]
    _fields = []



from ast import *

from ..util import CompilingNodeTransformer

"""
Make sure that certain variable names may never be overwritten
"""

FORBIDDEN_NAMES = {
    # type check
    "isinstance",
    # Type annotations
    "List",
    "Dict",
    "Union",
    # decorator and class name
    "dataclass",
    "PlutusData",
    # special decorator marking wrapped builtins
    "wraps_builtin",
}


class ForbiddenOverwriteError(ValueError):
    pass


class RewriteForbiddenOverwrites(CompilingNodeTransformer):
    step = "Checking for forbidden name overwrites"

    def visit_Name(self, node: Name) -> Name:
        if isinstance(node.ctx, Store) and node.id in FORBIDDEN_NAMES:
            raise ForbiddenOverwriteError(
                f"It is not allowed to overwrite name {node.id}"
            )
        return node
